# copyright ############################### #
# This file is part of the Xcoll Package.   #
# Copyright (c) CERN, 2025.                 #
# ######################################### #

import json
import time
import numpy as np

import xobjects as xo

try:
    from xaux import FsPath, ranID  # TODO: once xaux is in Xsuite keep only this
except (ImportError, ModuleNotFoundError):
    from ...xaux import FsPath, ranID

from .environment import format_fluka_float
from ...beam_elements.base import BaseCollimator
from ...materials import Material
from ...compare import deep_equal


class FlukaPrototype:
    # This is a registry to keep track of prototypes. If one is already defined,
    # we do not create a new instance but return the existing one
    _registry = []

    def __new__(cls, fedb_series=None, fedb_tag=None, **kwargs):
        _is_null = False
        # If the prototype is already defined, return the existing instance
        if fedb_series is None and fedb_tag is None:
            _is_null = True
        elif fedb_series is None or fedb_tag is None:
            raise ValueError("Both 'fedb_series' and 'fedb_tag' must be provided.")
        else:
            for prototype in FlukaPrototype._registry:
                if prototype.fedb_series.upper() == fedb_series.upper() \
                and prototype.fedb_tag.upper() == fedb_tag.upper() \
                and prototype.__class__ == cls:
                    return prototype
        # Register the new prototype
        self = object.__new__(cls)
        self._is_null = _is_null
        if not _is_null:
            FlukaPrototype._registry.append(self)
        return self

    def __init__(self, fedb_series=None, fedb_tag=None, *, angle=0, side=None, width=None,
                 height=None, length=None, material=None, info=None, extra_commands=None,
                 is_crystal=False, bending_radius=None, _allow_generic=False, is_broken=False,
                 _force_init=False, **kwargs):
        if getattr(self, "_initialized", False) and not _force_init:
            return
        self._idx = None
        self._type = self.__class__.__name__[5:].lower()
        self._file_is_valid = None
        # while True:
        #     self._hash = ranID(length=256)
        #     if self._hash not in [pp._hash for pp in FlukaPrototype._registry if pp is not self]:
        #         break
        if self._is_null:
            self._fedb_series = None
            self._fedb_tag = None
            self._name = None
            self._side = None
            self._angle = None
            self._length = None
            self._width = None
            self._height = None
            self._material = None
            self._is_crystal = None
            self._bending_radius = None
            self._info = None
            self._extra_commands = None
            self._is_broken = None
            self._initialized = True
            return
        if fedb_series == 'generic' and not _allow_generic:
            this_type = self.__class__.__name__[5:].lower()
            raise ValueError("Cannot use 'generic' as fedb_series, unless creating a generic " \
                          + f"{this_type}. Please use xcoll.fluka.create_generic_{this_type}() " \
                          + f"instead.")
        self._fedb_series = fedb_series
        self._fedb_tag = fedb_tag
        self._name = fedb_tag
        if side is not None:
            BaseCollimator.side.fset(self, side)  # This will overwrite the side in the FlukaCollimator
        else:
            self._side = None
        self._angle = angle        # This will adapt the transformation used in the LineBuilder
        self._length = length      # This will adapt the Drift used in pre-tracking (length_front, length_back)
        self._width = width
        self._height = height
        if isinstance(material, dict):
            if material.get('fluka_name') is not None \
            and material['fluka_name'].startswith('XCOLL'):
                # Remove fluka_name for previously auto-generated material,
                # such that the material code can be re-generated
                material.pop('fluka_name')
            material = Material.from_dict(material)
        self._material = material
        self._is_crystal = is_crystal
        self._bending_radius = bending_radius
        self._info = info
        self._extra_commands = extra_commands
        self._is_broken = is_broken
        self._generic_body_file = None
        self._generic_region_file = None
        self._generic_material_file = None
        self._initialized = True

    def __repr__(self):
        if self._is_null:
            return ''
        info = f" ({self.info})" if self.info else ''
        cry = "crystal " if self.is_crystal else ''
        defunct = " <defunct>" if self.is_defunct() else ''
        return f"{self.__class__.__name__} {cry}'{self.name}': " \
             + f"tag {self.fedb_tag} in {self.fedb_series} series{info}{defunct}"

    def __str__(self):
        if self._is_null:
            return ''
        return self.__repr__()

    # def __eq__(self, other):
    #     if not isinstance(other, FlukaPrototype):
    #         return False
    #     if self._is_null and other._is_null:
    #         return True
    #     if self._is_null != other._is_null:
    #         return False
    #     return deep_equal(self.to_dict(), other.to_dict())

    # def __hash__(self):
    #     return self._hash

    def is_defunct(self):
        if self._is_null:
            return False
        # Needed for proper handling of prototypes that are deleted
        return self not in self._registry

    def is_generic(self):
        if self._is_null:
            return False
        return self.fedb_series.lower() == 'generic'

    def delete(self, **kwargs):
        import xcoll as xc
        if self._is_null:
            return
        while self in self._registry:
            self._registry.remove(self)
        # Remove all files associated with the prototype
        for file in self.files:
            if file.exists() or file.is_symlink():
                try:
                    file.unlink()
                except FileNotFoundError:
                    pass
        fedb = xc.fluka.environment.fedb
        meta = fedb / "metadata" / f'{self.fedb_series}_{self.fedb_tag}.bodies.json'
        if meta.exists() or meta.is_symlink():
            try:
                meta.unlink()
            except FileNotFoundError:
                pass

    def to_dict(self):
        if self._is_null:
            return {'__class__': self.__class__.__name__}
        if self.is_defunct():
            raise ValueError(f"Cannot serialize defunct {self._type} '{self.name}'!")
        return {
            '__class__': self.__class__.__name__,
            'name': self.name,
            'fedb_series': self.fedb_series,
            'fedb_tag': self.fedb_tag,
            'side': self.side,
            'angle': self.angle,
            'length': self.length,
            'width': self.width,
            'height': self.height,
            'material': self.material.to_dict() if hasattr(self.material, 'to_dict') else self.material,
            'is_crystal': self.is_crystal,
            'bending_radius': self.bending_radius,
            'info': self.info,
            'extra_commands': self.extra_commands,
            'is_broken': self.is_broken,
        }

    @classmethod
    def from_dict(cls, data):
        cls = data.pop('__class__', None)
        if cls == 'FlukaPrototype':
            return FlukaPrototype(_allow_generic=True, _force_init=True, **data)
        elif cls == 'FlukaAssembly':
            return FlukaAssembly(_allow_generic=True, _force_init=True, **data)
        else:
            raise ValueError(f"Invalid data format for {cls}.")

    @classmethod
    def from_json(cls, path):
        if isinstance(path, str):
            path = FsPath(path)
        if not path.exists():
            raise FileNotFoundError(f"File {path} does not exist!")
        with path.open('r') as fid:
            data = json.load(fid)
        return cls.from_dict(data)

    @property
    def name(self):
        if self._is_null:
            return None
        return self._name

    @name.setter
    def name(self, val):
        if self._is_null:
            return
        if self.is_defunct():
            raise ValueError(f"Cannot rename defunct {self._type} '{self.name}'!")
        if len(val) > 8:
            raise ValueError(f"Prototype name '{val}' is too long (max. 8 characters).")
        self._name = val

    @property
    def fedb_series(self):
        if self._is_null:
            return None
        return self._fedb_series

    @property
    def fedb_tag(self):
        if self._is_null:
            return None
        return self._fedb_tag

    @property
    def body_file(self):
        import xcoll as xc
        if self._is_null:
            return None
        if self.is_generic():
            return self._generic_body_file
        fedb = xc.fluka.environment.fedb
        file = fedb / "bodies" / f"{self.fedb_series}_{self.fedb_tag}.bodies"
        return file.resolve()

    @body_file.setter
    def body_file(self, path):
        import xcoll as xc
        if self._is_null:
            raise ValueError("Cannot set body_file for a null prototype!")
        if self.is_defunct():
            raise ValueError(f"Cannot set body_file for defunct {self._type} '{self.name}'!")
        if self.is_generic():
            raise ValueError(f"Cannot set body_file for generic {self._type} '{self.name}'! "
                            + "These are generated automatically.")
        path = FsPath(path)
        if not path.exists():
            raise FileNotFoundError(f"File {path} does not exist!")
        fedb = xc.fluka.environment.fedb
        target = fedb / "bodies" / f"{self.fedb_series}_{self.fedb_tag}.bodies"
        if path != target:
            path.copy_to(target, method='mount')
        with open(fedb / "metadata" / f'{self.fedb_series}_{self.fedb_tag}.bodies.json', 'w') as fid:
            json.dump(self.to_dict(), fid, indent=4, cls=xo.JEncoder)

    @property
    def material_file(self):
        import xcoll as xc
        if self._is_null:
            return None
        if self.is_generic():
            return self._generic_material_file
        fedb = xc.fluka.environment.fedb
        file = fedb / "materials" / f"{self.fedb_series}_{self.fedb_tag}.assignmat"
        return file.resolve()

    @material_file.setter
    def material_file(self, path):
        import xcoll as xc
        if self._is_null:
            raise ValueError("Cannot set material_file for a null prototype!")
        if self.is_defunct():
            raise ValueError(f"Cannot set material_file for defunct {self._type} '{self.name}'!")
        if self.is_generic():
            raise ValueError(f"Cannot set material_file for generic {self._type} '{self.name}'! "
                            + "These are generated automatically.")
        path = FsPath(path)
        if not path.exists():
            raise FileNotFoundError(f"File {path} does not exist!")
        fedb = xc.fluka.environment.fedb
        path.copy_to(fedb / "materials" / f"{self.fedb_series}_{self.fedb_tag}.assignmat",
                     method='mount')

    @property
    def region_file(self):
        import xcoll as xc
        if self._is_null:
            return None
        if self.is_generic():
            return self._generic_region_file
        fedb = xc.fluka.environment.fedb
        file = fedb / "regions" / f"{self.fedb_series}_{self.fedb_tag}.regions"
        return file.resolve()

    @region_file.setter
    def region_file(self, path):
        import xcoll as xc
        if self._is_null:
            raise ValueError("Cannot set region_file for a null prototype!")
        if self.is_defunct():
            raise ValueError(f"Cannot set region_file for defunct {self._type} '{self.name}'!")
        if self.is_generic():
            raise ValueError(f"Cannot set region_file for generic {self._type} '{self.name}'! "
                            + "These are generated automatically.")
        path = FsPath(path)
        if not path.exists():
            raise FileNotFoundError(f"File {path} does not exist!")
        fedb = xc.fluka.environment.fedb
        path.copy_to(fedb / "regions" / f"{self.fedb_series}_{self.fedb_tag}.regions",
                     method='mount')

    def populate_into_temp_fedb(self, fedb):
        fedb = FsPath(fedb).resolve()
        if not fedb.exists():
            fedb.mkdir(parents=True)
            (fedb / 'assemblies').mkdir(parents=True)
            (fedb / 'bodies').mkdir(parents=True)
            (fedb / 'regions').mkdir(parents=True)
            (fedb / 'materials').mkdir(parents=True)
            (fedb / 'stepsizes').mkdir(parents=True)
        if self.is_generic():
            for f in self.files:
                if f is None or not f.exists():
                    raise ValueError(f"Generic prototype '{self.name}' is missing files!")
            return  # Generic prototypes are generated automatically
        link = fedb / 'bodies' / self.body_file.name
        if not link.exists():
            link.symlink_to(self.body_file)
        mat_link = fedb / 'materials' / f'{self.body_file.stem}.assignmat'
        if not mat_link.exists() and self.material_file is not None:
            mat_link.symlink_to(self.material_file)
        reg_link = fedb / 'regions' / f'{self.body_file.stem}.regions'
        if not reg_link.exists() and self.region_file is not None:
            reg_link.symlink_to(self.region_file)

    @property
    def files(self):
        return [self.body_file, self.material_file, self.region_file]

    def exists(self):
        if self.is_generic():
            return True
        if self._is_null or not self.files or self.is_defunct():
            return False
        return np.all([ff.exists() for ff in self.files])

    def assert_exists(self):
        if not self.exists():
            raise ValueError(f"{self._type.capitalize()} '{self.name}' "
                            + f"does not exist in the FEDB!")

    @property
    def side(self):
        if self._is_null:
            return None
        return BaseCollimator.side.fget(self)

    @property
    def angle(self):
        if self._is_null:
            return None
        return self._angle

    @property
    def length(self):
        if self._is_null:
            return None
        return self._length

    @property
    def width(self):
        if self._is_null:
            return None
        return self._width

    @property
    def height(self):
        if self._is_null:
            return None
        return self._height

    @property
    def material(self):
        if self._is_null:
            return None
        return self._material

    @property
    def is_crystal(self):
        if self._is_null:
            return None
        return self._is_crystal

    @property
    def bending_radius(self):
        if self._is_null:
            return None
        return self._bending_radius

    @property
    def info(self):
        if self._is_null:
            return None
        return self._info

    @property
    def extra_commands(self):
        if self._is_null:
            return None
        return self._extra_commands

    @property
    def is_broken(self):
        if self._is_null:
            return None
        return self._is_broken

    @property
    def fluka_position(self):
        if self._idx is None:
            return None
        # Maximum positions for the parking region:
        # x in [-3000.0, 3000.0], y in [-4000.0, -2000.0], z in [0.0, 1.E5]
        return 0., 0., 0., (self._idx%5-2)*500.0 , -3000., (self._idx//5+1)*1000.0

    def generate_code(self, idx=0, elements=[]):
        import xcoll as xc
        if self._is_null:
            raise ValueError(f"Cannot generate code for null {self._type}!")
        if self.is_defunct():
            raise ValueError(f"Cannot generate code for defunct {self._type} '{self.name}'!")
        if not self.exists():
            raise ValueError(f"{self._type.capitalize()} '{self.name}' "
                           + f"does not exist in the FEDB!")
        self.check_file_valid()
        _type = 'ASSEMBLY' if isinstance(self, FlukaAssembly) else 'PROTOTYPE'
        prot  = f"{_type:9}     {self.name}\n"
        prot += f"FEDB_SERIES   {self.fedb_series}\n"
        prot += f"FEDB_TAG      {self.fedb_tag}\n"
        prot += f"ROT-DEFI  "
        self._idx = idx  # Store the index for fluka_position property
        for value in self.fluka_position:
            prot += format_fluka_float(value)
        prot += " proto\n"
        if self._extra_commands:
            if hasattr(self._extra_commands, "__iter__") \
            and not isinstance(self._extra_commands, str):
                prot += "\n".join(self._extra_commands) + "\n"
            else:
                prot += self._extra_commands + "\n"
        maps = []
        for element in elements:
            if not hasattr(element, 'name'):
                raise ValueError(f"Element {element} has no `name` variable! "
                               + f"Cannot assign to a {self._type}!")
            if not isinstance(element, xc.fluka.engine._element_classes):
                raise ValueError(f"Element {element.name} is not a FLUKA element! "
                               + f"Cannot assign to a {self._type}!")
            if not element.assembly is self:
                raise ValueError(f"Element '{element.name}' is not assigned to "
                               + f"{self._type} '{self.name}'!")
            maps.append(f" {element.name.upper()}")
        if len(maps) > 0:
            prot_map = [f"MAP_ENTRIES  "]
            for element in maps:
                if len(prot_map[-1] + element) > 75:
                    prot_map.append(f"MAP_ENTRIES  ")
                prot_map[-1] += element
            prot += "\n".join(prot_map) + "\n"
        prot += '#'
        return prot

    def _check_name_clash(self, existing_names):
        if self.name.upper() in existing_names:
            i = 0
            while True:
                new_name = f"{self.name}{i}"
                if new_name.upper() not in existing_names:
                    print(f"Warning: {self._type.capitalize()} name {self.name} "
                        + f"already in use. Renaming to '{new_name}'.")
                    self.name = new_name
                    break
                i += 1
        existing_names.append(self.name.upper())

    @classmethod
    def _group_elements_by_prototype(cls, elements):
        import xcoll as xc
        # Sort elements by their assigned prototype/assembly
        all_prototypes = {}
        all_assemblies = {}
        if not hasattr(elements, '__iter__'):
            elements = [elements]
        for element in elements:
            if not isinstance(element, xc.fluka.engine._element_classes):
                raise ValueError(f"Element {element.name} is not a FLUKA element! "
                               + f"Cannot generate prototypes file!")
            assm = element.assembly
            if assm is None:
                raise ValueError(f"Element '{element.name}' has no assigned prototype! "
                               + f"Cannot generate prototypes file!")
            if isinstance(assm, FlukaAssembly):
                if assm not in all_assemblies:
                    all_assemblies[assm] = []
                all_assemblies[assm].append(element)
            elif isinstance(assm, FlukaPrototype):
                if assm not in all_prototypes:
                    all_prototypes[assm] = []
                all_prototypes[assm].append(element)
            else:
                raise ValueError(f"Element '{element.name}' has an invalid assigned prototype! "
                               + f"Cannot generate prototypes file!")
        return all_prototypes, all_assemblies

    @classmethod
    def make_prototypes_file(cls, elements, save=True, path=None):
        all_prototypes, all_assemblies = cls._group_elements_by_prototype(elements)
        # Generate code
        file_content = ["#...+....1....+....2....+....3....+....4....+....5....+....6....+....7....+....8"]
        i = 0  # Initialise i in case there are no prototypes
        existing_names = []
        # First the prototypes
        for i, (pro, els) in enumerate(all_prototypes.items()):
            assert isinstance(pro, FlukaPrototype)
            pro._check_name_clash(existing_names)
            file_content.append(pro.generate_code(idx=i, elements=els))
        # Then the assemblies
        for j, (assm, els) in enumerate(all_assemblies.items()):
            assert isinstance(assm, FlukaAssembly)
            assm._check_name_clash(existing_names)
            file_content.append(assm.generate_code(idx=i+j+1, elements=els))
        # Sanity check
        all_names = [assm.name for assm in (all_prototypes | all_assemblies).keys()]
        if len(all_names) != len(set(all_names)):
            raise ValueError("Duplicate prototype/assembly names found! "
                           + "Cannot generate prototypes file!")
        # Finalise
        file_content = "\n".join(file_content)
        if save:
            if path is None:
                path = FsPath.cwd()
            with (path / "prototypes.lbp").open("w") as fp:
                fp.write(file_content)
        return file_content

    @classmethod
    def inspect_prototypes_file(cls, elements, prototypes_file=None):
        if prototypes_file is None:
            prototypes_file = FsPath.cwd() / "prototypes.lbp"
        prototypes_in_file = {}
        with FsPath(prototypes_file).open("r") as fp:
            name = None
            for line in fp.readlines():
                line = line.strip()
                if line.startswith("#") or len(line) == 0:
                    continue
                if line.startswith("PROTOTYPE"):
                    name = line.split()[1]
                    prototypes_in_file[name] = {'type': 'PROTOTYPE'}
                elif line.startswith("ASSEMBLY"):
                    name = line.split()[1]
                    prototypes_in_file[name] = {'type': 'ASSEMBLY'}
                elif line.startswith("FEDB_SERIES"):
                    if name is None:
                        raise ValueError(f"File {prototypes_file} malformed ("
                                    + "FEDB_SERIES without PROTOTYPE or ASSEMBLY)!")
                    prototypes_in_file[name]['fedb_series'] = line.split()[1]
                elif line.startswith("FEDB_TAG"):
                    if name is None:
                        raise ValueError(f"File {prototypes_file} malformed ("
                                    + "FEDB_TAG without PROTOTYPE or ASSEMBLY)!")
                    prototypes_in_file[name]['fedb_tag'] = line.split()[1]
                elif line.startswith("MAP_ENTRIES"):
                    if name is None:
                        raise ValueError(f"File {prototypes_file} malformed ("
                                    + "MAP_ENTRIES without PROTOTYPE or ASSEMBLY)!")
                    if 'elements' not in prototypes_in_file[name]:
                        prototypes_in_file[name]['elements'] = []
                    prototypes_in_file[name]['elements'] += line.split()[1:]
        prototypes, assemblies = cls._group_elements_by_prototype(elements)
        for prototype, els in (prototypes | assemblies).items():
            if prototype.name not in prototypes_in_file:
                raise ValueError(f"Prototype {prototype.name} (in {prototype.fedb_series}) not found in prototypes file!")
            else:
                _type = 'ASSEMBLY' if isinstance(prototype, FlukaAssembly) else 'PROTOTYPE'
                if _type != prototypes_in_file[prototype.name]['type']:
                    raise ValueError(f"Wrong type for {prototype.name} (in {prototype.fedb_series}) (expected "
                                  + f"{prototype._type.upper()}, got "
                                  + f"{prototypes_in_file[prototype.name]['type']})!")
                if 'fedb_series' not in prototypes_in_file[prototype.name]:
                    raise ValueError(f"FEDB_SERIES for {prototype.name} (in {prototype.fedb_series}) not found in prototypes file!")
                if prototype.fedb_series != prototypes_in_file[prototype.name]['fedb_series']:
                    raise ValueError(f"Wrong fedb_series for {prototype.name} (in {prototype.fedb_series}) (expected "
                                   + f"{prototype.fedb_series}, got "
                                   + f"{prototypes_in_file[prototype.name]['fedb_series']})!")
                if 'fedb_tag' not in prototypes_in_file[prototype.name]:
                    raise ValueError(f"FEDB_TAG for {prototype.name} (in {prototype.fedb_series}) not found in prototypes file!")
                if prototype.fedb_tag != prototypes_in_file[prototype.name]['fedb_tag']:
                    raise ValueError(f"Wrong fedb_tag for {prototype.name} (in {prototype.fedb_series}) (expected "
                                   + f"{prototype.fedb_tag}, got "
                                   + f"{prototypes_in_file[prototype.name]['fedb_tag']})!")
                if 'elements' not in prototypes_in_file[prototype.name]:
                    raise ValueError(f"MAP_ENTRIES for {prototype.name} (in {prototype.fedb_series}) not found in prototypes file!")
                for element in els:
                    if element.name.upper() not in prototypes_in_file[prototype.name]['elements']:
                        raise ValueError(f"Element {element.name} not found in prototypes file!")

    @property
    def dependant_assemblies(self):
        """Return a list of all assemblies that depend on this prototype."""
        if self._is_null:
            return []
        assemblies = []
        for assembly in FlukaPrototype._registry:
            if isinstance(assembly, FlukaAssembly) \
            and assembly.fedb_series == self.fedb_series:
                if self in assembly.prototypes:
                    assemblies.append(assembly)
        return assemblies

    def check_file_valid(self, raise_error=True):
        return True

    def view(self, show=True, keep_files=False):
        import xcoll as xc
        if self.exists():
            xc.fluka.environment.test_assembly(self.fedb_series, self.fedb_tag, show=show,
                                               keep_files=keep_files)


class FlukaAssembly(FlukaPrototype):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._generic_assembly_file = None

    def delete(self, _ignore_files=False, **kwargs):
        import xcoll as xc
        if self._is_null:
            return
        # Remove prototypes if no other assembly depends on them
        if self.assembly_file.exists() or not _ignore_files:
            try:
                to_delete = []  # Need to do like this to avoid prototypes spawning again (during checking of dependant assemblies)
                for pro in self.prototypes:
                    if pro.dependant_assemblies == [self]:
                        to_delete.append(pro)
                for pro in to_delete:
                    # Only delete the prototype if no other assembly depends on it
                    try:
                        pro.delete()
                    except Exception as e:
                        print(f"Could not remove prototype '{pro.name}': {e}")
            except Exception as e:
                print(f"Could not remove dependent prototypes: {e}")
        # Remove the assembly from the registry of all prototypes
        while self in FlukaPrototype._registry:
            FlukaPrototype._registry.remove(self)
        # Remove all files associated with the assembly
        if self.assembly_file.exists() or self.assembly_file.is_symlink():
            try:
                self.assembly_file.unlink()
            except FileNotFoundError:
                pass
        fedb = xc.fluka.environment.fedb
        meta = fedb / "metadata" / f'{self.fedb_series}_{self.fedb_tag}.lbp.json'
        if meta.exists() or meta.is_symlink():
            meta.unlink()

    @property
    def assembly_file(self):
        import xcoll as xc
        if self._is_null:
            return None
        if self.is_generic():
            return self._generic_assembly_file
        fedb = xc.fluka.environment.fedb
        file = fedb / "assemblies" / f"{self.fedb_series}_{self.fedb_tag}.lbp"
        return file.resolve()

    @assembly_file.setter
    def assembly_file(self, path):
        import xcoll as xc
        if self._is_null:
            raise ValueError("Cannot set assembly_file for a null assembly!")
        if self.is_defunct():
            raise ValueError(f"Cannot set assembly_file for defunct assembly '{self.name}'!")
        if self.is_generic():
            raise ValueError(f"Cannot set assembly_file for generic assembly '{self.name}'!\n"
                           + f"It is generated automatically.")
        path = FsPath(path)
        if not path.exists():
            raise FileNotFoundError(f"File {path} does not exist!")
        fedb = xc.fluka.environment.fedb
        target = fedb / "assemblies" / f"{self.fedb_series}_{self.fedb_tag}.lbp"
        if path != target:
            path.copy_to(target, method='mount')
        with open(fedb / "metadata" / f'{self.fedb_series}_{self.fedb_tag}.lbp.json', 'w') as fid:
            dct = self.to_dict()
            # Remove fluka_name for previously auto-generated material,
            # such that the material code can be re-generated
            if 'material' in dct and dct['material'] is not None and 'fluka_name' in dct['material'] \
            and dct['material']['fluka_name'] is not None and dct['material']['fluka_name'].startswith('XCOLL'):
                dct['material'].pop('fluka_name')
            json.dump(dct, fid, indent=4, cls=xo.JEncoder)

    @property
    def body_file(self):
        pass

    @body_file.setter
    def body_file(self, path):
        pass

    @property
    def material_file(self):
        pass

    @material_file.setter
    def material_file(self, path):
        pass

    @property
    def region_file(self):
        pass

    @region_file.setter
    def region_file(self, path):
        pass

    def populate_into_temp_fedb(self, fedb):
        fedb = FsPath(fedb).resolve()
        if not fedb.exists():
            raise ValueError(f"FEDB path {fedb} does not exist!")
        if self.is_generic():
            from xcoll.scattering_routines.fluka.generic_prototype import (
                _assembly_file, _body_file, _crystal_body_file, _region_file,
                _crystal_region_file, _material_file, _crystal_material_file
            )
            if self.is_crystal:
                body_file, tank_file = _crystal_body_file(fedb, self.fedb_tag,
                    self.length, self.bending_radius, self.width, self.height)
                body_region_file, tank_region_file = _crystal_region_file(fedb, self.fedb_tag)
                body_mat_file, tank_mat_file = _crystal_material_file(fedb, self.fedb_tag, self.material)
            else:
                body_file, tank_file = _body_file(fedb, self.fedb_tag, self.length,
                                                  self.width, self.height)
                body_region_file, tank_region_file = _region_file(fedb, self.fedb_tag)
                body_mat_file, tank_mat_file = _material_file(fedb, self.fedb_tag, self.material)
            for pro in self.prototypes:
                if pro.name.endswith('_B'):
                    pro._generic_body_file = body_file
                    pro._generic_region_file = body_region_file
                    pro._generic_material_file = body_mat_file
                elif pro.name.endswith('_T'):
                    pro._generic_body_file = tank_file
                    pro._generic_region_file = tank_region_file
                    pro._generic_material_file = tank_mat_file
                else:
                    raise ValueError(f"Generic assembly prototype '{pro.name}' has invalid name! "
                                   + "Expected to end with '_B' or '_T'.")
            self._generic_assembly_file = _assembly_file(fedb, self.fedb_tag, self.side)
        else:
            link = fedb / 'assemblies' / self.assembly_file.name
            if not link.exists():
                link.symlink_to(self.assembly_file)
        for prot in self.prototypes:
            prot.populate_into_temp_fedb(fedb)

    @property
    def prototypes(self):
        if not hasattr(self, '_prototypes'):
            if not self.assembly_file.exists():
                raise FileNotFoundError(f"Assembly file {self.assembly_file} not found!")
            prototypes = []
            prototype_found = False
            fedb_series = None
            fedb_tag = None
            # Try maximally 5 times to parse the assembly file (in case it's being written by another process)
            for i in range(5):
                success = True
                with self.assembly_file.open('r') as fid:
                    for line in fid:
                        if line.upper().startswith('PROTOTYPE'):
                            prototype_found = True
                            continue
                        if line.upper().startswith('FEDB_SERIES'):
                            if not prototype_found:
                                if i < 4:
                                    # Retry parsing the file
                                    success = False
                                    break
                                raise ValueError("Corrupt assembly file: FEDB_SERIES without PROTOTYPE.")
                            fedb_series = line.split()[1]
                        if line.upper().startswith('FEDB_TAG'):
                            if not prototype_found:
                                if i < 4:
                                    # Retry parsing the file
                                    success = False
                                    break
                                raise ValueError("Corrupt assembly file: FEDB_TAG without PROTOTYPE.")
                            fedb_tag = line.split()[1]
                        if fedb_series is not None and fedb_tag is not None:
                            prototypes.append(FlukaPrototype(fedb_series=fedb_series, fedb_tag=fedb_tag,
                                                            _allow_generic=True))
                            fedb_series = None
                            fedb_tag = None
                            prototype_found = False
                if success:
                    break
                else:
                    time.sleep(0.1)
            if prototype_found:
                raise ValueError("Corrupt assembly file: incomplete prototype definition.")
            self._prototypes = prototypes
        return self._prototypes

    @property
    def files(self):
        if self.assembly_file.exists():
            files = [self.assembly_file]
            for prot in self.prototypes:
                files += prot.files
            return files

    def check_file_valid(self, raise_error=True):
        if self._file_is_valid is None or self._file_is_valid == False:
            if self.assembly_file is None:
                self._file_is_valid = False
                if raise_error:
                    raise ValueError("Assembly has no assembly_file defined!")
                return self._file_is_valid
            self._file_is_valid = True
            prototype_found = False
            fedb_series = None
            fedb_tag = None
            ass_fedb_tag = None
            # Try maximally 5 times to parse the assembly file (in case it's being written by another process)
            for i in range(5):
                success = True
                with self.assembly_file.open('r') as fid:
                    for line in fid:
                        if line.upper().startswith('ASSEMBLY'):
                            if ass_fedb_tag:
                                self._file_is_valid = False
                                if raise_error:
                                    if i < 4:
                                        # Retry parsing the file
                                        success = False
                                        break
                                    raise ValueError("Corrupt assembly file: ASSEMBLY defined more than once.")
                            ass_fedb_tag = line.split()[1]
                            continue
                        if line.upper().startswith('PROTOTYPE'):
                            if ass_fedb_tag:
                                self._file_is_valid = False
                                if raise_error:
                                    if i < 4:
                                        # Retry parsing the file
                                        success = False
                                        break
                                    raise ValueError("Corrupt assembly file: PROTOTYPE after ASSEMBLY.")
                            prototype_found = True
                            continue
                        if line.upper().startswith('FEDB_SERIES'):
                            if fedb_series is not None:
                                self._file_is_valid = False
                                if raise_error:
                                    if i < 4:
                                        # Retry parsing the file
                                        success = False
                                        break
                                    raise ValueError("Corrupt assembly file: FEDB_SERIES defined more than once.")
                            if not prototype_found:
                                self._file_is_valid = False
                                if raise_error:
                                    if i < 4:
                                        # Retry parsing the file
                                        success = False
                                        break
                                    raise ValueError("Corrupt assembly file: FEDB_SERIES without PROTOTYPE.")
                            fedb_series = line.split()[1]
                        if line.upper().startswith('FEDB_TAG'):
                            if fedb_tag is not None:
                                self._file_is_valid = False
                                if raise_error:
                                    if i < 4:
                                        # Retry parsing the file
                                        success = False
                                        break
                                    raise ValueError("Corrupt assembly file: FEDB_TAG defined more than once.")
                            if not prototype_found:
                                self._file_is_valid = False
                                if raise_error:
                                    if i < 4:
                                        # Retry parsing the file
                                        success = False
                                        break
                                    raise ValueError("Corrupt assembly file: FEDB_TAG without PROTOTYPE.")
                            fedb_tag = line.split()[1]
                        if fedb_series is not None and fedb_tag is not None:
                            fedb_series = None
                            fedb_tag = None
                            prototype_found = False
                if not ass_fedb_tag:
                    self._file_is_valid = False
                    if raise_error:
                        if i < 4:
                            # Retry parsing the file
                            success = False
                        else:
                            raise ValueError("Corrupt assembly file: ASSEMBLY not defined.")
                if ass_fedb_tag != self.fedb_tag:
                    self._file_is_valid = False
                    if raise_error:
                        if i < 4:
                            # Retry parsing the file
                            success = False
                        else:
                            raise ValueError(f"Corrupt assembly file: ASSEMBLY {ass_fedb_tag} "
                                    + f"does not match {self.fedb_tag}. Please take note "
                                    + f"that the filename should match exactly, i.e. "
                                    + f"{self.fedb_series}_{self.fedb_tag}.lbp")
                if success:
                    break
                else:
                    time.sleep(0.1)
        return self._file_is_valid


class FlukaPrototypeAccessor:
    """This class is used to access the prototypes in the FEDB."""

    def __init__(self):
        self._type = 'Prototype'

    @property
    def _raw(self):
        return [(pro.fedb_series.lower(), pro.fedb_tag.lower(), pro)
                for pro in FlukaPrototype._registry
                if not isinstance(pro, FlukaAssembly)]

    def __repr__(self):
        return f"<FlukaPrototypeAccessor at {hex(id(self))} (use .show() to see the content)>"

    def __str__(self):
        res = []
        for ss, vv in self.ordered_data.items():
            res.append(f"FEDB Series: {ss}")
            for tt, pro in vv.items():
                if pro.is_broken:
                    continue
                res.append(f"    {tt:<16}: {pro}")
        res_broken = []
        for ss, vv in self.ordered_data.items():
            if any(pro.is_broken for pro in vv.values()):
                res_broken.append(f"FEDB Series: {ss}")
            for tt, pro in vv.items():
                if not pro.is_broken:
                    continue
                res_broken.append(f"    {tt:<16}: {pro}")
        if res_broken:
            res.append('')
            res.append('BROKEN:')
            res.append('')
            res += res_broken
        return "\n".join(res)

    def show(self):
        """Print the content of the FEDB."""
        print(self)

    @property
    def series(self):
        return {ss for ss, _, _ in self._raw}

    @property
    def tags(self):
        return {tt for _, tt, _ in self._raw}

    @property
    def data(self):
        return {f'{ss}_{tt}': pro for ss, tt, pro in self._raw}

    @property
    def ordered_data(self):
        return {
            ss: {tt: vv for xx, tt, vv in self._raw if xx == ss}
            for ss in self.series
        }

    def keys(self):
        f"""A set-like object providing a view on the {self._type.lower()} tags."""
        return self.data.keys()

    def values(self):
        f"""A set-like object providing a view on the {self._type.lower()[-1]}ies."""
        return self.data.values()

    def items(self):
        f"""A set-like object providing a view on the {self._type.lower()[-1]}ies and their tags."""
        return self.data.items()

    def __len__(self):
        return len(self.data)

    def __iter__(self):
        return iter(self.values().__iter__())

    def __iter__(self):
        super().__setattr__('_iter_data', iter(self.keys()))
        return self

    def __next__(self):
        try:
            name = next(self._iter_data)
        except StopIteration:
            raise StopIteration
        else:
            return self[name]

    def __contains__(self, val):
        val = val.lower()
        return val in self.data or val in self.series or val in self.tags

    def __getitem__(self, val):
        val = val.lower()
        # If val is the full specification (series_tag), return the prototype
        if val in self.data:
            return self.data[val]
        # If val is a series, return the prototypes in that series
        elif val in self.series or val == 'generic':   # Special case for Andre
            return FlukaSeriesAccessor(self._type, self.ordered_data, val)
        # If val is a tag and that tag is unique, return the prototype
        elif val in self.tags:
            this_data = [pro for ss, tt, pro in self._raw if tt == val]
            if len(this_data) > 1:
                raise KeyError(f"Tag '{val}' is not unique in the FEDB. "
                              + "Please specify the series as well.")
            else:
                return this_data[0]
        else:
            raise KeyError(f"{self._type} '{val}' not found in the FEDB.")


class FlukaAssemblyAccessor(FlukaPrototypeAccessor):
    """This class is used to access the assemblies in the FEDB."""

    def __init__(self):
        self._type = 'Assembly'

    @property
    def _raw(self):
        return [(pro.fedb_series.lower(), pro.fedb_tag.lower(), pro)
                for pro in FlukaPrototype._registry
                if isinstance(pro, FlukaAssembly)]


class FlukaSeriesAccessor:
    """This class is used to access the prototypes or assemblies in a specific series."""

    def __init__(self, type, data, series):
        self._type = type
        self._series = series
        if series == 'generic' and series not in data:
            self._series_data = {}
        else:
            self._series_data = data[series]

    def __repr__(self):
        return f"<FlukaSeriesAccessor at {hex(id(self))} (use .show() to see the content)>"

    def __str__(self):
        res = [f"FEDB Series: {self._series}"]
        for tt, pro in self._series_data.items():
            if pro.is_broken:
                continue
            res.append(f"    {tt:<16}: {pro}")
        res_broken = []
        for tt, pro in self._series_data.items():
            if not pro.is_broken:
                continue
            res_broken.append(f"    {tt:<16}: {pro}")
        if res_broken:
            res.append('')
            res.append('BROKEN:')
            res.append('')
            res += res_broken
        return "\n".join(res)

    def show(self):
        """Print the content of the FEDB series."""
        print(self)

    def keys(self):
        f"""A set-like object providing a view on the {self._type.lower()} tags."""
        return self._series_data.keys()

    def values(self):
        f"""A set-like object providing a view on the {self._type.lower()[-1]}ies."""
        return self._series_data.values()

    def items(self):
        f"""A set-like object providing a view on the {self._type.lower()[-1]}ies and their tags."""
        return self._series_data.items()

    def __len__(self):
        return len(self._series_data)

    def __iter__(self):
        return iter(self.values().__iter__())

    def __contains__(self, val):
        val = val.lower()
        return val in self._series_data

    def __getitem__(self, val):
        val = val.lower()
        # If val is a tag, return the prototype
        if val in self._series_data:
            return self._series_data[val]
        raise KeyError(f"{self._type} with tag '{val}' not found in series "
                     + f"'{self._series}'.")



# copyright ############################### #
# This file is part of the Xcoll package.   #
# Copyright (c) CERN, 2025.                 #
# ######################################### #

from .material import Material
from .atoms import Beryllium, Aluminium, Silicon, Copper, Germanium, Molybdenum, Tungsten, Lead
from .allotropes import CarbonFibreCarbon
from .mixtures import MolybdenumGraphite, CopperDiamond, Glidcop15, Inermet180
from .database import _manually_add_material_to_db


# K2 variants of elements
K2Beryllium  = Beryllium.adapt(A=9.01, radiation_length=0.353)
K2Beryllium._num_nucleons_eff = 3.3668216738035692

K2Aluminium  = Aluminium.adapt(A=26.98, radiation_length=0.089)
K2Aluminium._num_nucleons_eff = 4.852801185429141

K2Silicon    = Silicon.adapt(A=28.08, density=2.33, radiation_length=1)
K2Silicon._num_nucleons_eff = 4.917875746143141

K2Copper     = Copper.adapt(A=63.55, radiation_length=0.0143)
K2Copper._num_nucleons_eff = 6.456795558713786

K2Germanium  = Germanium.adapt(radiation_length=1)
K2Germanium._num_nucleons_eff = 6.750726791090729

K2Molybdenum = Molybdenum.adapt(A=95.96, radiation_length=0.0096, density=10.22)
K2Molybdenum._num_nucleons_eff = 7.4075614639458625

K2Tungsten   = Tungsten.adapt(A=183.85, radiation_length=0.0035)
K2Tungsten._num_nucleons_eff = 9.200252118694163

K2Lead       = Lead.adapt(density=11.35, radiation_length=0.0056)
K2Lead._num_nucleons_eff = 9.57417689618882



# K2 variants of allotropes
K2CarbonFibreCarbon = CarbonFibreCarbon.adapt(A=12.01, radiation_length=0.2557)
K2CarbonFibreCarbon._num_nucleons_eff = 3.705323974123165

K2Carbon2 = K2CarbonFibreCarbon.adapt(density=4.52, A=12.01, radiation_length=0.094)  # Was in SixTrack, not sure what it is
K2Carbon2._num_nucleons_eff = 3.705323974123165


# K2 variants of compounds
K2MolybdenumGraphite = MolybdenumGraphite.adapt(excitation_energy=87.1, radiation_length=0.1193, density=2.5)
K2MolybdenumGraphite._ZA_mean = 0.4915003695491501
K2MolybdenumGraphite._Z2_eff = 44.222500000000004
K2MolybdenumGraphite._num_nucleons_eff = 3.8554740361466595
K2MolybdenumGraphite._atoms_per_volume = 1.1127384996304508e+29

K2CopperDiamond = CopperDiamond.adapt(excitation_energy=152.9, radiation_length=0.0316)
K2CopperDiamond._ZA_mean = 0.47147385103011097
K2CopperDiamond._Z2_eff = 141.61
K2CopperDiamond._num_nucleons_eff = 4.74615190534514
K2CopperDiamond._atoms_per_volume = 1.2884136332805071e+29

K2Glidcop15 = Glidcop15.adapt(excitation_energy=320.8, radiation_length=0.0144)
K2Glidcop15._ZA_mean = 0.45605700712589076
K2Glidcop15._Z2_eff = 829.44
K2Glidcop15._num_nucleons_eff = 6.4432201272142064
K2Glidcop15._atoms_per_volume = 8.515869673285828e+28

K2Inermet180 = Inermet180.adapt(excitation_energy=682.2, radiation_length=0.00385)
K2Inermet180._ZA_mean = 0.40611877624475107
K2Inermet180._Z2_eff = 4583.29
K2Inermet180._num_nucleons_eff = 8.904790721425886
K2Inermet180._atoms_per_volume = 6.5026114985003005e+28


_manually_add_material_to_db(K2Beryllium,          'K2Beryllium')
_manually_add_material_to_db(K2Aluminium,          'K2Aluminium')
_manually_add_material_to_db(K2Silicon,            'K2Silicon')
_manually_add_material_to_db(K2Copper,             'K2Copper')
_manually_add_material_to_db(K2Germanium,          'K2Germanium')
_manually_add_material_to_db(K2Molybdenum,         'K2Molybdenum')
_manually_add_material_to_db(K2Tungsten,           'K2Tungsten')
_manually_add_material_to_db(K2Lead,               'K2Lead')
_manually_add_material_to_db(K2CarbonFibreCarbon,  'K2CarbonFibreCarbon')
_manually_add_material_to_db(K2Carbon2,            'K2Carbon2')
_manually_add_material_to_db(K2MolybdenumGraphite, 'K2MolybdenumGraphite')
_manually_add_material_to_db(K2CopperDiamond,      'K2CopperDiamond')
_manually_add_material_to_db(K2Glidcop15,          'K2Glidcop15')
_manually_add_material_to_db(K2Inermet180,         'K2Inermet180')


# Clean up namespace
del (Beryllium, Aluminium, Silicon, Copper, Germanium, Molybdenum, Tungsten, Lead,
     CarbonFibreCarbon, MolybdenumGraphite, CopperDiamond, Glidcop15, Inermet180)
del Material



# copyright ############################### #
# This file is part of the Xcoll Package.   #
# Copyright (c) CERN, 2025.                 #
# ######################################### #

import xtrack.particles.pdg as pdg

class BaseWrapper:
    _engine_cls = None
    _environment_cls = None
    _particle_masses_meta = None
    _particle_names_meta = None

    def __init__(self):
        self._engine = None
        self._environment = None

    def __str__(self):
        return f"{self.__class__.__name__}"

    def __repr__(self):
        return f"<{str(self)} at {hex(id(self))}>"

    @property
    def engine(self):
        self._lazy_load_engine()
        return self._engine

    @property
    def environment(self):
        self._lazy_load_environment()
        return self._environment

    @property
    def particle_masses(self):
        if self._particle_masses_meta is None:
            raise NotImplementedError(
                "This wrapper does not implement particle_masses.")
        return BaseWrapperAccessor(self._particle_masses_meta)

    @property
    def particle_names(self):
        if self._particle_names_meta is None:
            raise NotImplementedError(
                "This wrapper does not implement particle_names.")
        return BaseWrapperAccessor(self._particle_names_meta)

    def _lazy_load_environment(self):
        # Ensure the environment is loaded
        if not self._environment:
            self._environment = self._environment_cls()

    def _lazy_load_engine(self):
        # Ensure the engine is loaded
        if not self._engine:
            self._engine = self._engine_cls()
            self._engine._environment = self.environment
            if self._particle_masses_meta is None:
                self._engine._masses = None
            else:
                self._engine._masses = self.particle_masses


class BaseWrapperAccessor:
    """Accessor to get reference masses."""

    def __init__(self, meta_dict):
        self._meta_dict = meta_dict

    def __getitem__(self, pdgid: int | str) -> float | None:
        if isinstance(pdgid, str):
            pdgid = pdg.get_pdg_id_from_name(pdgid)
        vals = [vv['value'] for _, vv in self._meta_dict.items()
                if vv['pdg_id'] == pdgid]
        if len(vals) == 1:
            return vals[0]
        vals = [vv['value'] for _, vv in self._meta_dict.items()
                if vv['pdg_id'] == -pdgid]
        if len(vals) == 1:
            return vals[0]

    def __contains__(self, pdgid: int | str) -> bool:
        return self[pdgid] is not None


# copyright ############################### #
# This file is part of the Xcoll Package.   #
# Copyright (c) CERN, 2025.                 #
# ######################################### #

import os
import json
import numpy as np
from subprocess import run, PIPE
from contextlib import redirect_stdout

from ...beam_elements import FlukaCrystal
from ...beam_elements.base import OPEN_GAP, OPEN_JAW
from .prototype import FlukaPrototype
from .includes import get_include_files

try:
    from xaux import FsPath  # TODO: once xaux is in Xsuite keep only this
except (ImportError, ModuleNotFoundError):
    from ...xaux import FsPath

_header_start = "*  XCOLL START  **"
_header_stop  = "*  XCOLL END  **"

def _stop_and_error(old_cwd, error):
    if old_cwd is not None:
        os.chdir(old_cwd)
    import xcoll as xc
    xc.fluka.engine.stop()
    raise error

def create_fluka_input(element_dict, particle_ref, prototypes_file=None,
                       verbose=True, cwd=None, **kwargs):
    import xcoll as xc
    old_cwd = None
    if cwd is not None:
        old_cwd = FsPath.cwd()
        os.chdir(cwd)
    assemblies = []
    for nn, ee in element_dict.items():
        if isinstance(ee, FlukaCrystal):
            if not ee.assembly.is_crystal:
                _stop_and_error(old_cwd, ValueError(f"Collimator {nn} has a crystal element but its assembly is not a crystal!"))
        assemblies.append(ee.assembly)
    try:
        kwargs['assemblies'] = assemblies
        for assm in assemblies:
            if assm.material is not None:
                if assm.material.fluka_name is None or assm.material.fluka_name.startswith('XCOLL'):
                    assm.material._generate_fluka_code()
        fedb = xc.fluka.environment.create_temp_fedb(assemblies)
        _create_prototypes_file(element_dict, prototypes_file)
        _, kwargs = get_include_files(particle_ref, verbose=verbose, **kwargs)
    except Exception as e:
        _stop_and_error(old_cwd, e)
    # Call FLUKA_builder
    try:
        collimator_dict = _element_dict_to_fluka(element_dict)
        input_file, fluka_dict = _fluka_builder(collimator_dict, fedb=fedb)
    except Exception as e:
        _stop_and_error(old_cwd, e)
    input_file = FsPath(input_file).resolve()
    insertion_file = (input_file.parent / 'insertion.txt').resolve()
    if not input_file.exists() or not insertion_file.exists():
        _stop_and_error(old_cwd, FileNotFoundError("LineBuilder did not create the expected output files!"))
    # Expand using include files
    cmd = run([(fedb / 'tools' / 'expand.sh').as_posix(), input_file.name],
              cwd=FsPath.cwd(), stdout=PIPE, stderr=PIPE)
    if cmd.returncode == 0:
        if verbose:
            print("Expanded include files.")
    else:
        stderr = cmd.stderr.decode('UTF-8').strip().split('\n')
        _stop_and_error(old_cwd, RuntimeError(f"Could not expand include files!\nError given is:\n{stderr}"))
    new_input_file = input_file.parent / f'{input_file.stem}_exp.inp'
    if not new_input_file.exists():
        _stop_and_error(old_cwd, FileNotFoundError("expand.sh did not create the expected expanded input file!"))
    input_file.rename(input_file.parent / f'{input_file.stem}_orig.inp')
    new_input_file.rename(input_file)
    try:
        _write_xcoll_header_to_fluka_input(input_file, fluka_dict, element_dict, verbose)
    except Exception as e:
        _stop_and_error(old_cwd, e)
    if cwd is not None:
        os.chdir(old_cwd)
    if verbose:
        print(f"Created FLUKA input file {input_file}.")
    return [input_file, insertion_file], kwargs


def get_collimators_from_input_file(input_file):
    with open(input_file, 'r') as fp:
        data = fp.read()
    if _header_start not in data or _header_stop not in data:
        raise ValueError("No Xcoll header found in input file. Regenerate input file!")
    commented_dict = data.split(_header_start)[1].split(_header_stop)[0].split('\n')[1:-1]
    cleaned_dict = "".join([val[3:] for val in commented_dict])
    return json.loads(cleaned_dict)


def verify_insertion_file(insertion_file, element_dict):
    all_fluka_ids = []
    with open(insertion_file, 'r') as fid:
        for line in fid.readlines():
            all_fluka_ids.append(int(line.split()[0]))
    for name, val in element_dict.items():
        if val.fluka_id not in all_fluka_ids:
            raise ValueError(f'FlukaCollimator {name} not found in insertion file!')


def _create_prototypes_file(element_dict, prototypes_file=None):
    for name, ee in element_dict.items():
        if ee.assembly is None:
            raise ValueError(f"Collimator {name} has no assembly!")
    if prototypes_file is None:
        FlukaPrototype.make_prototypes_file(element_dict.values())
    else:
        prototypes_file = FsPath(prototypes_file).resolve()
        prototypes_file.copy_to(FsPath.cwd() / 'prototypes.lbp', method='mount')
    FlukaPrototype.inspect_prototypes_file(element_dict.values(), prototypes_file)


# TODO check that prototype is valid and its sides
def _element_dict_to_fluka(element_dict, dump=False):
    collimator_dict = {}
    for name, ee in element_dict.items():
        if ee.length < 1.e-12:
            raise ValueError(f"Collimator {name} has zero length!")

        nsig = 1 # TODO can remove?
        if ee.assembly.is_crystal:
            if ee.jaw is None:
                half_gap = OPEN_JAW
            else:
                nsig = ee.gap
                half_gap = -ee.jaw  # Mistake in LineBuilder?
            offset = 0
            tilt_1 = 0
            tilt_2 = ee.tilt
        else:
            tilt_1 = ee.tilt_L
            tilt_2 = ee.tilt_R
            if ee.side == 'left':
                if ee.jaw_L is None:
                    half_gap = OPEN_JAW
                else:
                    nsig = ee.gap_L
                    half_gap = ee.jaw_L
                offset = 0
                tilt_2 = 0
            elif ee.side == 'right':
                if ee.jaw_R is None:
                    half_gap = OPEN_JAW
                else:
                    nsig = ee.gap_R
                    half_gap = -ee.jaw_R   #  TODO: is the sign correct?
                offset = 0
                tilt_1 = 0
            else:
                if ee.jaw_L is None and ee.jaw_R is None:
                    half_gap = OPEN_JAW
                    offset = 0
                else:
                    if ee.gap_L is not None:
                        nsig = ee.gap_L
                    elif ee.gap_R is not None:
                        nsig = ee.gap_R
                    half_gap = (ee._jaw_LU + ee._jaw_LD - ee._jaw_RU - ee._jaw_RD) / 4
                    offset   = (ee._jaw_LU + ee._jaw_LD + ee._jaw_RU + ee._jaw_RD) / 4
        tilt_1 = round(tilt_1, 9)
        tilt_2 = round(tilt_2, 9)

        if nsig is None:
            nsig = 1

        collimator_dict[name] = {
            'name': name,
            'betx': 1,
            'bety': 1,
            'material': 'stub',
            'length': ee.length,
            'angle': round(np.deg2rad(ee.angle) - ee.assembly.angle, 9),
            'sigma_x': 1,
            'sigma_y': 1,
            'offset': offset,
            'tilt_1': tilt_1,
            'tilt_2': tilt_2,
            'nsig': nsig,
            'half_gap': half_gap
        }
    if dump:
        # dump coll_dictionary in json format
        with open('collimator_dict.json', 'w') as fp:
            json.dump(collimator_dict, fp, indent=4)
    return collimator_dict


def _fluka_builder(collimator_dict, fedb):
    import xcoll as xc
    # Save system state
    xc.fluka.environment.set_fedb_environment(fedb)
    file_path = xc.fluka.environment.linebuilder / "src" / "FLUKA_builder.py"
    if file_path.exists():
        try:
            import FLUKA_builder as fb
        except ImportError as e:
            raise EnvironmentError(f"Cannot import FLUKA_builder: {e}")
    else:
        raise EnvironmentError(f"FLUKA_builder.py not found at: {file_path.as_posix()}")
    collimatorList = fb.CollimatorList()
    collimatorList.acquireCollxsuite(collimator_dict)

    args_fb = fb.args_fluka_builder()
    args_fb.collimatorList = collimatorList
    args_fb.geometrical_emittance = None
    args_fb.prototype_file = 'prototypes.lbp'
    args_fb.output_name = 'fluka_input'
    args_fb.fedb_u_path = fedb.as_posix()
    import importlib, structure
    importlib.reload(structure)
    with open('linebuilder.log', 'w') as f:
        with redirect_stdout(f):
            input_file, coll_dict = fb.fluka_builder(args_fb, auto_accept=True)

    # Restore system state
    xc.fluka.environment.restore_environment()

    return input_file, coll_dict


def _write_xcoll_header_to_fluka_input(input_file, fluka_dict, element_dict, verbose):
    # Check that all collimators were treated and write header
    for name, ee in element_dict.items():
        if name.lower() not in fluka_dict:
            if verbose:
                print(f"Warning: Collimator {name} was requested but not treated by "
                    + f"the LineBuilder.")
        else:
            fluka_dict[name.lower()]['length'] /= 100
            fluka_dict[name.lower()]['angle'] = ee.angle
            if ee.assembly.is_crystal:
                if ee.side == 'left':
                    fluka_dict[name.lower()]['tilt'] = [ee.tilt, None]
                    fluka_dict[name.lower()]['jaw'] = [ee.jaw, None]
                elif ee.side == 'right':
                    fluka_dict[name.lower()]['tilt'] = [None, ee.tilt]
                    fluka_dict[name.lower()]['jaw'] = [None, ee.jaw]
            else:
                fluka_dict[name.lower()]['tilt'] = [ee.tilt_L, ee.tilt_R]
                fluka_dict[name.lower()]['jaw'] = [ee.jaw_L, ee.jaw_R]
    header = ["*  DO NOT CHANGE THIS HEADER", _header_start, "*  {"]
    for kk, vv in fluka_dict.items():
        header.append(f'*  "{kk}": ' + json.dumps(vv).replace('"jaw"', '\n*          "jaw"') + ',')
    header[-1] = header[-1][:-1]  # remove last comma
    header.append("*  }")
    header.append(_header_stop)

    with open(input_file, 'r') as fp:
        data = fp.read()
    with open(input_file, 'w') as fp:
        fp.write("\n".join(header) + "\n*\n" + data)



# copyright ############################### #
# This file is part of the Xcoll package.   #
# Copyright (c) CERN, 2025.                 #
# ######################################### #

from .material import Material
from .database import db, _manually_add_material_to_db
from .atoms import (Carbon, Oxygen, Argon, Hydrogen, Boron, Nitrogen, Silicon,
                    Aluminium, Iron, Nickel, Cobalt, Zinc, Chromium, Titanium,
                    Zirconium, Manganese, Phosphorus, Sulfur, Copper, Molybdenum,
                    Tungsten, Calcium, Niobium, Magnesium, Tantalum)
from .compounds import CarbonDioxide, BoronNitride, Silica, BoronTrioxide, AluminiumOxide, NiZnFerrite


# TODO: add more from Geant4: https://geant4-userdoc.web.cern.ch/UsersGuides/ForApplicationDeveloper/html/Appendix/materialNames.html


# Mixture definitions from FLUKA
# ===============================

# By molar fractions (like number of atoms but not a compound)
CopperDiamond      = Material(components=[Copper, Carbon, Chromium, Zirconium], molar_fractions=[0.2696, 0.7269, 0.0033, 0.0002], density=5.4, state='solid')
TiZrMo             = Material(components=[Molybdenum, Carbon, Titanium, Zirconium], molar_fractions=[0.98762, 0.00158, 0.00996, 0.00084], density=10.22, state='solid')
MolybdenumGraphite = Material(components=[Molybdenum, Carbon, Titanium], molar_fractions=[0.0201, 0.978, 0.0019], density=2.57, state='solid')
MG6403Fc           = Material(components=[Molybdenum, Carbon, Titanium], molar_fractions=[0.0184, 0.9809, 0.0007], density=2.55, state='solid')
MG6400             = Material(components=[Molybdenum, Carbon], molar_fractions=[0.0146, 0.9854], density=2.48, state='solid')

MolybdenumGraphite.info = 'Molybdenum-Graphite composite (NB-8404Ng), used for LHC TCPPM and TCSPM.'
MG6403Fc.info           = 'Molybdenum-Graphite composite. Candidate material for LHC collimators, but only used for TCPPMP prototype.'
MG6400.info             = 'Molybdenum-Graphite composite. Candidate material for LHC collimators, but not used.'

# By mass fractions
Air                 = Material(components=[Nitrogen, Oxygen, Argon, CarbonDioxide], state='gas', pressure=1, temperature=288.15,
                               mass_fractions=[78.084, 20.946, 0.934, 0.036], density=0.001225)
BoronNitride5000    = Material(components=[BoronNitride, Silica, Oxygen, BoronTrioxide, Calcium], state='solid',
                               mass_fractions=[0.929, 0.002, 0.04, 0.004, 0.025], density=1.925)
FeNiCo              = Material(components=[Iron, Nickel, Cobalt], state='solid',
                               mass_fractions=[0.54, 0.29, 0.17], density=8.35)
Glidcop15           = Material(components=[Copper, AluminiumOxide], state='solid',
                               mass_fractions=[0.9985, 0.0015], density=8.93)
CuNiFeMn            = Material(components=[Copper, Nickel, Iron, Manganese], state='solid',
                               mass_fractions=[0.875, 0.1, 0.015, 0.01], density=8.87)
Inermet180          = Material(components=[Tungsten, Nickel, Copper], state='solid',
                               mass_fractions=[0.95, 0.035, 0.015], density=18.0)
StainLessSteel304L  = Material(components=[Iron, Chromium, Nickel, Manganese, Silicon, Phosphorus, Sulfur, Carbon], state='solid',
                               mass_fractions=[0.67145, 0.185, 0.1125, 0.02, 0.01, 0.00045, 0.0003, 0.0003], density=8.02)
StainLessSteel316L  = Material(components=[Iron, Chromium, Nickel, Molybdenum, Manganese, Silicon, Phosphorus, Sulfur, Carbon], state='solid',
                               mass_fractions=[0.61395, 0.185, 0.14, 0.03, 0.02, 0.01, 0.00045, 0.0003, 0.0003], density=8.0)
StainLessSteel316LN = Material(components=[Iron, Chromium, Nickel, Molybdenum, Manganese, Silicon, Titanium, Nitrogen, Niobium, Copper, Cobalt, Phosphorus, Carbon, Sulfur, Tantalum, Boron], state='solid',
                               mass_fractions=[0.65093, 0.17, 0.12, 0.025, 0.02, 0.0075, 0.0015, 0.0014, 0.001, 0.001, 0.0005, 0.00045, 0.0003, 0.0003, 0.0001, 0.00002], density=8.03)
TitaniumNitride     = Material(components=[Titanium, Nitrogen], state='solid',
                               mass_fractions=[0.77, 0.23], density=5.22)
Antico              = Material(components=[Aluminium, Silicon, Iron, Magnesium, Manganese, Chromium, Titanium, Zinc, Copper], state='solid',
                               mass_fractions=[0.9725, 0.006, 0.005, 0.004, 0.004, 0.0035, 0.002, 0.002, 0.001], density=2.7)
CopperChromium      = Material(components=[Copper, Chromium], state='solid',
                               mass_fractions=[0.991, 0.009], density=8.8)
CastIron            = Material(components=[Iron, Carbon, Silicon], state='solid',
                               mass_fractions=[0.94, 0.04, 0.02], density=7.2)
CuCrZr              = Material(components=[Copper, Chromium, Iron, Silicon, Zirconium], state='solid',
                               mass_fractions=[0.98805, 0.0085, 0.0008, 0.001, 0.00165],  density=8.9)
TiGr2               = Material(components=[Titanium, Iron, Carbon, Oxygen, Nitrogen, Hydrogen], state='solid',
                               mass_fractions=[0.99325, 0.003, 0.0008, 0.0025, 0.0003, 0.00015], density=4.51)


# Extra parameters for Everest
# ============================

MolybdenumGraphite.adapt(inplace=True, nuclear_radius=0.25, nuclear_elastic_slope=76.7,
                         cross_section=[0.362, 0.247, 0, 0, 0, 0.0094e-2], hcut=0.02)
CopperDiamond.adapt(inplace=True, nuclear_radius=0.308, nuclear_elastic_slope=115.0,
                    cross_section=[0.572, 0.370, 0, 0, 0, 0.0279e-2], hcut=0.02)
Glidcop15.adapt(inplace=True, nuclear_radius=0.418, nuclear_elastic_slope=208.7,
                cross_section=[1.246, 0.765, 0, 0, 0, 0.1390e-2], hcut=0.02)
Inermet180.adapt(inplace=True, nuclear_radius=0.578, nuclear_elastic_slope=392.1,
                 cross_section=[2.548, 1.473, 0, 0, 0, 0.5740e-2], hcut=0.02)


# Existing FLUKA mixtures
# =======================

# POLYSTYR Polystyrene 1.06
# PLASCINT Plastic scintillator 1.032
# PMMA Polymethyl methacrylate, Plexiglas, Lucite, Perspex 1.19
# BONECOMP Compact bone 1.85
# BONECORT Cortical bone 1.85
# MUSCLESK Skeletal muscle 1.04
# MUSCLEST Striated muscle 1.04
# ADTISSUE Adipose tissue 0.92
# KAPTON Kapton polyimide film 1.42
# POLYETHY Polyethylene 0.94


# Metadata for database
# =====================

_manually_add_material_to_db(CopperDiamond,       'CopperDiamond',       short_name='CuCD', fluka_name='CUDIAM75')
_manually_add_material_to_db(TiZrMo,              'TiZrMo')
_manually_add_material_to_db(MolybdenumGraphite,  'MolybdenumGraphite',  short_name='MoGR')
_manually_add_material_to_db(MG6403Fc,            'MG6403Fc',                               fluka_name='MG6403Fc')
_manually_add_material_to_db(MG6400,              'MG6400',                                 fluka_name='MoGRMG64')
_manually_add_material_to_db(Air,                 'Air',                                    fluka_name='AIR')
_manually_add_material_to_db(BoronNitride5000,    'BoronNitride5000',                       fluka_name='BN5000')
_manually_add_material_to_db(FeNiCo,              'FeNiCo',                                 fluka_name='ASTMF-15')
_manually_add_material_to_db(Glidcop15,           'Glidcop15',           short_name='Glid', fluka_name='GLIDCP15')
_manually_add_material_to_db(CuNiFeMn,            'CuNiFeMn',                               fluka_name='CuNiFeMn')
_manually_add_material_to_db(Inermet180,          'Inermet180',          short_name='Iner', fluka_name='INERM180')
_manually_add_material_to_db(StainLessSteel304L,  'StainLessSteel304L',                     fluka_name='SS304L')
_manually_add_material_to_db(StainLessSteel316L,  'StainLessSteel316L',                     fluka_name='SS316L')
_manually_add_material_to_db(StainLessSteel316LN, 'StainLessSteel316LN',                    fluka_name='SS316LN')
_manually_add_material_to_db(TitaniumNitride,     'TitaniumNitride',                        fluka_name='TITNAT')
_manually_add_material_to_db(Antico,              'Antico',                                 fluka_name='ANTICO')
_manually_add_material_to_db(CopperChromium,      'CopperChromium',                         fluka_name='CuCr')
_manually_add_material_to_db(CastIron,            'CastIron',                               fluka_name='CASTIRON')
_manually_add_material_to_db(CuCrZr,              'CuCrZr',                                 fluka_name='CuCrZr')
_manually_add_material_to_db(TiGr2,               'TiGr2',                                  fluka_name='TiGr2')


# Clean up namespace
del (Carbon, Oxygen, Argon, Hydrogen, Boron, Nitrogen, Silicon,
     Aluminium, Iron, Nickel, Cobalt, Zinc, Chromium, Titanium,
     Zirconium, Manganese, Phosphorus, Sulfur, Copper, Molybdenum,
     Tungsten, Calcium, Niobium, Magnesium, Tantalum, CarbonDioxide,
     BoronNitride, Silica, BoronTrioxide, AluminiumOxide, NiZnFerrite)
del Material
del db



# copyright ############################### #
# This file is part of the Xcoll package.   #
# Copyright (c) CERN, 2025.                 #
# ######################################### #

import os
import re
import sys
from typing import Iterable, Tuple, Union

# ---------- ANSI basics ----------
_RESET = "\033[0m"

def _sgr(*codes: Union[int, str]) -> str:
    return f"\033[{';'.join(str(c) for c in codes)}m"

# Text attributes (SGR)
_SGR_CODES = {
    "bold": 1,
    "dim": 2,
    "italic": 3,
    "underline": 4,
}

# ANSI named colours: normal + bright
# You can extend this dict as you wish.
_NAMED = {
    "black": 0, "red": 1, "green": 2, "yellow": 3, "blue": 4, "magenta": 5, "cyan": 6, "white": 7,
    "bright_black": 8, "bright_red": 9, "bright_green": 10, "bright_yellow": 11,
    "bright_blue": 12, "bright_magenta": 13, "bright_cyan": 14, "bright_white": 15,
}

# Add friendly aliases mapped to 256-colour indices (xterm palette)
# (These go through the 38;5;<index> pathway.)
_ALIAS_256 = {
    # ---- Greys (common names) ----
    "grey": 8, "gray": 8,
    "dark_grey": 240, "dark_gray": 240,
    "slate_grey": 243, "slate_gray": 243,
    "light_grey": 250, "light_gray": 250,
    "gainsboro": 252,
    "silver": 251,
    "whitesmoke": 255,

    # Fine greyscale steps (darker -> lighter)
    "grey_0": 232, "grey_1": 233, "grey_2": 234, "grey_3": 235,
    "grey_4": 236, "grey_5": 237, "grey_6": 238, "grey_7": 239,
    "grey_8": 240, "grey_9": 241, "grey_10": 242, "grey_11": 243,
    "grey_12": 244, "grey_13": 245, "grey_14": 246, "grey_15": 247,
    "grey_16": 248, "grey_17": 249, "grey_18": 250, "grey_19": 251,
    "grey_20": 252, "grey_21": 253, "grey_22": 254, "grey_23": 255,
    "gray_0": 232, "gray_1": 233, "gray_2": 234, "gray_3": 235,
    "gray_4": 236, "gray_5": 237, "gray_6": 238, "gray_7": 239,
    "gray_8": 240, "gray_9": 241, "gray_10": 242, "gray_11": 243,
    "gray_12": 244, "gray_13": 245, "gray_14": 246, "gray_15": 247,
    "gray_16": 248, "gray_17": 249, "gray_18": 250, "gray_19": 251,
    "gray_20": 252, "gray_21": 253, "gray_22": 254, "gray_23": 255,

    # ---- Reds ----
    "maroon": 88, "firebrick": 124, "crimson": 160, "red2": 196,
    "tomato": 203, "salmon": 210, "light_coral": 210, "indian_red": 167,
    "rosy_brown": 138, "brown": 94,

    # ---- Oranges / Browns ----
    "dark_orange": 166, "orange": 208, "coral": 209, "sandy_brown": 215,
    "peru": 173, "chocolate": 166, "sienna": 130, "tan": 180,
    "burlywood": 180, "peach": 216,

    # ---- Yellows / Golds ----
    "gold": 220, "goldenrod": 178, "dark_khaki": 143, "khaki": 186,
    "lemon": 226, "light_yellow": 229, "wheat": 229, "cornsilk": 230,

    # ---- Greens ----
    "dark_green": 22, "forest_green": 22, "olive": 58, "olive_drab": 64,
    "green4": 28, "sea_green": 29, "dark_sea_green": 108, "spring_green": 48,
    "lawn_green": 118, "chartreuse": 118, "pale_green": 120,
    "medium_spring_green": 49, "yellow_green": 190, "mint_green": 121,

    # ---- Teals / Cyans ----
    "teal": 30, "dark_cyan": 30, "cadet_blue": 109, "light_sea_green": 37,
    "turquoise": 45, "medium_turquoise": 44, "aquamarine": 86,
    "cyan2": 51, "light_cyan": 195,

    # ---- Blues ----
    "navy": 17, "midnight_blue": 17, "dark_blue": 18, "blue2": 21,
    "royal_blue": 63, "dodger_blue": 33, "deep_sky_blue": 39,
    "cornflower_blue": 69, "steel_blue": 67, "light_steel_blue": 147,
    "sky_blue": 117,

    # ---- Purples / Violets ----
    "indigo": 54, "dark_violet": 92, "blue_violet": 93,
    "slate_blue": 62, "medium_purple": 104,
    "orchid": 170, "plum": 176, "violet": 177, "thistle": 182,

    # ---- Pinks / Magentas ----
    "magenta2": 201, "fuchsia": 201, "deep_pink": 199,
    "hot_pink": 205, "pink": 211, "light_pink": 218, "pale_violet_red": 168,

    # ---- Earth / Neutrals ----
    "beige": 230, "antique_white": 230, "linen": 255, "ivory": 255,
}



# Regex for stripping ANSI
_ANSI_RE = re.compile(r"\x1B\[[0-?]*[ -/]*[@-~]")

def strip_ansi(s: str) -> str:
    """Remove ANSI escape codes from a string."""
    return _ANSI_RE.sub("", s)

def pad_styled(s: str, width: int, *, align: str = 'left') -> str:
    pad = width + len(s) - len(strip_ansi(s))
    if align == 'left':
        return f"{s:{pad}}"
    elif align == 'right':
        return f"{s:>{pad}}"
    elif align == 'center':
        return f"{s:^{pad}}"
    raise ValueError(f"Unknown alignment: {align}")


# ---------- Colour encoders ----------
def _encode_colour(
    colour: Union[str, int, Tuple[int, int, int], None], *, bg: bool = False
) -> Iterable[int]:
    """
    Return SGR code sequence (as ints) for a colour.

    Accepts:
      - Named colour: "red", "bright_blue", ...
      - 256-colour index: int 0..255
      - True-colour: (r,g,b) 0..255
      - Hex string: "#RRGGBB" or "RRGGBB"
      - None: no colour
    """
    if colour is None:
        return ()

    if isinstance(colour, str):
        s = colour.lower().strip()
        if s.startswith("#"):
            s = s[1:]
        if len(s) == 6 and all(c in "0123456789abcdef" for c in s):
            r = int(s[0:2], 16); g = int(s[2:4], 16); b = int(s[4:6], 16)
            return (48 if bg else 38, 2, r, g, b)
        if s in _NAMED:
            idx = _NAMED[s]
            # map 0..7 to 30..37 (fg) / 40..47 (bg), 8..15 to 90..97 / 100..107
            if idx < 8:
                base = 40 if bg else 30
                return (base + idx,)
            else:
                base = 100 if bg else 90
                return (base + (idx - 8),)
        if s in _ALIAS_256:
            return (48 if bg else 38, 5, _ALIAS_256[s])
        raise ValueError(f"Unknown colour name: {colour!r}")

    elif isinstance(colour, int):
        if not (0 <= colour <= 255):
            raise ValueError("256-colour index must be in 0..255")
        return (48 if bg else 38, 5, colour)

    elif isinstance(colour, tuple) and len(colour) == 3:
        r, g, b = colour
        for c in (r, g, b):
            if not (0 <= int(c) <= 255):
                raise ValueError("RGB components must be in 0..255")
        return (48 if bg else 38, 2, int(r), int(g), int(b))

    raise TypeError("Colour must be str|int|tuple[int,int,int]|None")


# ---------- Style core ----------
def style(
    text: str,
    *,
    color: Union[str, int, Tuple[int, int, int], None] = None,
    colour: Union[str, int, Tuple[int, int, int], None] = None,
    bg: Union[str, int, Tuple[int, int, int], None] = None,
    bold: bool = False,
    italic: bool = False,
    underline: bool = False,
    dim: bool = False,
    enabled: bool | None = None,
    reset: bool = True,
) -> str:
    """
    Wrap `text` with ANSI styles.

    - colo(u)r/bg: named ("red", "bright_blue"), 0..255, (r,g,b), or "#RRGGBB"
    - toggles: bold, italic, underline, dim
    - enabled: override auto-detection. If None, enable only when stdout is a TTY and NO_COLOR not set.
    - reset: append a trailing reset (recommended when printing standalone fragments)
    """
    if enabled is None:
        enabled = sys.stdout.isatty() and "NO_COLOR" not in os.environ

    if not enabled:
        return text

    codes: list[int] = []
    if colour is not None:
        if color is not None:
            raise ValueError("Cannot specify both 'color' and 'colour'")
        color = colour
    codes += list(_encode_colour(color, bg=False))
    codes += list(_encode_colour(bg, bg=True))
    for flag, name in ((bold, "bold"), (dim, "dim"), (italic, "italic"), (underline, "underline")):
        if flag:
            codes.append(_SGR_CODES[name])

    if not codes:
        return text

    prefix = _sgr(*codes)
    return f"{prefix}{text}{_RESET if reset else ''}"


def stylize_chunks(chunks: Iterable[str], *, reset_between: bool = False) -> str:
    """
    Concatenate pre-styled fragments safely. If your fragments already end with RESET,
    set reset_between=False (default). If they do NOT, set reset_between=True to avoid
    style bleed between fragments.
    """
    if not reset_between:
        return "".join(chunks)
    out = []
    for ch in chunks:
        out.append(ch)
        if not ch.endswith(_RESET):
            out.append(_RESET)
    return "".join(out)


# # ---------- Example integration with your class ----------
# class MaterialsDatabase:
#     def __init__(self, entries=42):
#         self.entries = entries

#     def _plain(self) -> str:
#         return f"MaterialsDatabase with {self.entries} entries"

#     def _pretty(self) -> str:
#         name = style("MaterialsDatabase", bold=True)
#         count = style(str(self.entries), color="bright_cyan", italic=True)
#         return f"{name} with {count} entries"

#     def __str__(self) -> str:
#         # Always plain for logging / files / f-strings
#         return self._plain()

#     def show(self) -> None:
#         # Pretty for terminal output
#         print(self._pretty())



# copyright ############################### #
# This file is part of the Xcoll package.   #
# Copyright (c) CERN, 2025.                 #
# ######################################### #

from __future__ import annotations

import sys
from enum import Enum
from warnings import warn
from types import MappingProxyType, ModuleType
from typing import Dict, Mapping, Optional, Union, Tuple, Type, List, Set


ConstantType = Union[int, float, bool, str]


__all__ = ["constant", "group", "Constants"]


# This code implements a metaclass-based system for defining constants with
# metadata, ensuring uniqueness and providing C header generation capabilities.
#
# The constants will automatically be exported to the defining module and can be
# merged into other modules. They also support grouping and can generate C
# header files with include guards. The C-names of the constants are verified to
# be unique globally (to avoid collisions in C code), while the python names are
# verified to be unique across all modules they are exported to. Additionally,
# if specified with the keyword 'unique', the values of the constants are also
# verified to be unique globally (and not only across the modules), as this might
# be required for reverse lookups in the C code.


# ---- global registries (cross-module uniqueness) ----------------------------
_XO_CONST_GLOBAL_TYPE_REG: Dict[str, List[Type[Constants]]] = {}    # category -> [defining classes]
_XO_CONST_GLOBAL_NAME_REG: Dict[str, Set[str]] = {}                 # module -> NAMEs
_XO_CONST_GLOBAL_CNAME_REG: Dict[str, Tuple[str,str]] = {}          # c_name -> (NAME, module)
_XO_CONST_GLOBAL_UNIQUE_VALUE_REG: Dict[str, Dict[Union[int, float, str], Tuple[str, str]]] = {}  # category -> value -> (NAME, module)
_XO_CONST_GLOBAL_COUNT_REG: Dict[str, int] = {}  # For unnamed constants in groups and include guards in C source

# ---- helpers ----------------------------------------------------------------
def _cjoin(prefix: str, name: str, sep: str = "_") -> str:
    if not prefix:
        return name
    return f"{prefix}{'' if prefix.endswith(sep) else sep}{name}"

def _pluralise(category: str, explicit: Optional[str]) -> str:
    if explicit:
        return explicit
    n = category.lower()
    if n.endswith(("s", "x", "z", "ch", "sh")): return category + "es"
    if len(category) >= 2 and n.endswith("y") and category[-2] not in "aeiou": return category[:-1] + "ies"
    return category + "s"

def _to_number(val: ConstantType) -> Union[int, float]:
    if isinstance(val, bool):
        return int(val)           # for C and numeric maps
    if isinstance(val, int):
        return val
    if isinstance(val, float):
        return val
    raise TypeError(f"Unsupported constant type: {type(val).__name__}")

def _to_value_key(val: ConstantType) -> Union[int, float, str]:
    if isinstance(val, str):
        return val
    return _to_number(val)

def _to_c_literal(val: ConstantType) -> str:
    if isinstance(val, str):
        # Escape backslashes and quotes for a C string literal
        s = val.replace("\\", "\\\\").replace('"', r'\"')
        return f"\"{s}\""
    # numeric/bool -> numeric literal
    return str(_to_number(val))

def _to_c_header(meta, include_guard, contents: Optional[List[str]] = None) -> str:
    kk = f"{include_guard}_guard"
    if kk not in _XO_CONST_GLOBAL_COUNT_REG:
        _XO_CONST_GLOBAL_COUNT_REG[kk] = -1
    _XO_CONST_GLOBAL_COUNT_REG[kk] += 1
    include_guard += f"_I{_XO_CONST_GLOBAL_COUNT_REG[kk]}"
    lines = [f"#ifndef {include_guard}",
            f"#define {include_guard}"]
    if contents:
        lines.extend(contents)
    else:
        for vv in meta.values():     # already in definition order
            c_rhs = _to_c_literal(vv["value"])
            lines.append(f"  #define {vv['c_name']:35} {c_rhs:25}     // {vv['info']}")
    lines.append(f"#endif /* {include_guard} */")
    lines.append("")
    return "\n".join(lines)

def _merge_map(mod: ModuleType, attr: str, new_items: Mapping,
               extend_values_when_duplicate: bool = False) -> None:
    existing = getattr(mod, attr, None)
    if existing is None:
        merged = dict(new_items)  # preserve insertion
    elif isinstance(existing, ModuleType):
        raise TypeError(f"Cannot merge '{attr}' in {mod.__name__}: existing is a module, not a mapping. "
                         "Do you have a python file named like the attribute?")
    else:
        merged = dict(existing)
        for nk, nv in new_items.items():
            if nk in merged and merged[nk] != nv:
                if extend_values_when_duplicate:
                    merged[nk] = tuple(dict.fromkeys(merged[nk] + nv))  # merge groups, preserve order
                else:
                    raise ValueError(f"Duplicate {attr}[{nk!r}] in {mod.__name__}: {merged[nk]} vs {nv}")
            else:
                merged[nk] = nv
    setattr(mod, attr, MappingProxyType(merged))

def _merge_c_header(mod: ModuleType, attr: str, src: str, include_guard: str) -> None:
    if hasattr(mod, attr):
        # Append to existing source
        mod_src = getattr(mod, attr).splitlines()
        new_src = ['  ' + ll for ll in src.splitlines()]
        # We need a wrapper around the individual sources to avoid clashes
        if len(mod_src) > 2 and mod_src[2].startswith(f"  #ifndef"):
            # Wrapper exists already
            new_src = "\n".join([*mod_src[:-1], *new_src, mod_src[-1], ''])
        else:
            mod_src = ['  ' + ll for ll in mod_src]
            new_src = _to_c_header(None, include_guard, contents=[*mod_src, *new_src])
        setattr(mod, attr, new_src)
    else:
        setattr(mod, attr, src)

def _check_uniqueness(const_name, mod_name):
    if mod_name not in _XO_CONST_GLOBAL_NAME_REG:
        _XO_CONST_GLOBAL_NAME_REG[mod_name] = set()
    if const_name in _XO_CONST_GLOBAL_NAME_REG[mod_name]:
        raise ValueError(f"Constant name {const_name!r} already defined in module {mod_name}; uniqueness required.")
    _XO_CONST_GLOBAL_NAME_REG[mod_name].add(const_name)

def _register_type(cls, category, plural, reverse):
    # Register type (for cross-module consistency checks)
    if category in _XO_CONST_GLOBAL_TYPE_REG:
        for other_cls in _XO_CONST_GLOBAL_TYPE_REG[category]:
            if other_cls._plural_ != plural:
                raise ValueError(f"Category {category!r} already defined with plural "
                            + f"{other_cls._plural_!r} by {other_cls.__module__}"
                            + f".{other_cls.__name__}. Please make _plural_ consistent.")
            if other_cls._reverse_ != reverse:
                raise ValueError(f"Category {category!r} already defined with reverse "
                            + f"{other_cls._reverse_!r} by {other_cls.__module__}"
                            + f".{other_cls.__name__}. Please make _reverse_ consistent.")
        _XO_CONST_GLOBAL_TYPE_REG[category].append(cls)
    else:
        _XO_CONST_GLOBAL_TYPE_REG[category] = [cls]

def _check_cname_collisions(own_specs, mod_name):
    # Check for C name collisions
    for nm, spec in own_specs.items():
        cn = spec.c_name
        if cn in _XO_CONST_GLOBAL_CNAME_REG and _XO_CONST_GLOBAL_CNAME_REG[cn] != (nm, mod_name):
            other = _XO_CONST_GLOBAL_CNAME_REG[cn]
            raise ValueError(f"C macro {cn!r} collides between {other} and {(nm, mod_name)}")
        _XO_CONST_GLOBAL_CNAME_REG[cn] = (nm, mod_name)

def _check_value_uniqueness(own_specs, category, mod_name):
    if category not in _XO_CONST_GLOBAL_UNIQUE_VALUE_REG:
        _XO_CONST_GLOBAL_UNIQUE_VALUE_REG[category] = {}
    for const_name, spec in own_specs.items():
        vkey = _to_value_key(spec.py_value)
        if vkey in _XO_CONST_GLOBAL_UNIQUE_VALUE_REG[category]:
            other_name, other_mod = _XO_CONST_GLOBAL_UNIQUE_VALUE_REG[category][vkey]
            raise ValueError(
                f"Value {vkey!r} for {const_name!r} in {mod_name} "
                f"already used by {other_name!r} in {other_mod}; "
                f"values must be globally unique when _reverse_ == 'unique'."
            )
    for const_name, spec in own_specs.items():
        _XO_CONST_GLOBAL_UNIQUE_VALUE_REG[category][_to_value_key(spec.py_value)] = (const_name, mod_name)

def _get_specs(ns, c_prefix):
    # Collect specifications in definition order
    own_specs: Dict[str, ConstantSpec] = {}
    own_group_specs: Dict[str, GroupSpec] = {}
    for k, v in ns.items():
        if k.startswith("_"):
            continue
        if isinstance(v, ConstantSpec):
            v.name = k  # Remember constant name for lookup in groups (potentially in other classes)
            if v.c_name is None:
                v.c_name = _cjoin(c_prefix, k)
            own_specs[k] = v
        elif isinstance(v, ConstantType):
            own_specs[k] = ConstantSpec(v, info ='', c_name=_cjoin(c_prefix, k), name=k)
        elif isinstance(v, GroupSpec):
            v.name = k  # Remember group name for lookup in other classes
            own_group_specs[k] = v
        else:
            continue
    return own_specs, own_group_specs

def _resolve_group_specs(own_group_specs, category, mod_name, name_to_val):
    for gname, spec in own_group_specs.items():
        seq = []
        names = []
        for nm in spec.names:
            # References to other groups
            if isinstance(nm, GroupSpec):       # Reference to group spec
                if nm.name in own_group_specs:  # Defined in this class, so take resolved values
                    seq.extend(own_group_specs[nm.name].py_values)
                    names.extend(own_group_specs[nm.name].names)
                else:                           # Defined in other class, take original spec values
                    seq.extend(nm.py_values)
                    names.extend(nm.names)
                continue
            elif isinstance(nm, _TupleMixin):   # Reference to group defined in other classes
                seq.extend(nm.value)
                names.extend(nm.names)
                continue
            elif isinstance(nm, ConstantSpec):  # Reference to constant spec
                if isinstance(nm, str):
                    raise TypeError(f"Group {gname!r} in {mod_name} contains a string literal; groups are numeric-only.")
                if nm.name in name_to_val:      # Defined in this class, so take resolved value
                    seq.append(name_to_val[nm.name])
                    names.append(nm.name)
                else:                           # Defined in other class, take original spec value
                    seq.append(nm.py_value)
                    names.append(nm.name)
                continue
            elif isinstance(nm, (_IntMixin, _FloatMixin)):  # Reference to constant defined in other classes
                seq.append(nm.value)
                names.append(nm.name)
                continue
            elif isinstance(nm, str):           # String reference to constant or group defined in this class
                if nm in own_group_specs:
                    seq.extend(own_group_specs[nm].py_values)
                    names.extend(own_group_specs[nm].names)
                    continue
                elif nm in name_to_val:
                    seq.append(name_to_val[nm])
                    names.append(nm)
                    continue
            elif isinstance(nm, ConstantType):  # Reference got evaluated already - named constant lost
                if isinstance(nm, str):
                    raise TypeError(f"Group {gname!r} in {mod_name} contains a string literal; groups are numeric-only.")
                seq.append(_to_number(nm))
                kk = f"{category}_unnamed"
                if kk not in _XO_CONST_GLOBAL_COUNT_REG:
                    _XO_CONST_GLOBAL_COUNT_REG[kk] = -1
                _XO_CONST_GLOBAL_COUNT_REG[kk] += 1
                new_name = f"{category.upper()}_UNNAMED_{_XO_CONST_GLOBAL_COUNT_REG[kk]}"
                names.append(new_name)
                warn(f"Group {gname!r} in {mod_name} contains literal {nm!r}. "
                    + f"Name lost, assigned {new_name!r}.", UserWarning)
                continue
            raise KeyError(f"Group {gname!r} references unknown constant {nm!r}")
        # Ensure no duplicates and sort
        name_to_value = {}
        for vv, nn in zip(seq, names):
            if nn in name_to_value:
                if name_to_value[nn] != vv:
                    raise ValueError(f"Constant {nn!r} appears with two different "
                                    + f"values: {name_to_value[nn]} and {vv}")
            else:
                name_to_value[nn] = vv
        sorted_items = sorted(name_to_value.items(), key=lambda kv: kv[1])
        names = tuple(ll for ll, _ in sorted_items)
        seq   = tuple(vv for _, vv in sorted_items)
        own_group_specs[gname] = GroupSpec(names=names, py_values=seq, info=spec.info, name=gname)


# ---- spec holders used in class bodies --------------------------------------
class ConstantSpec:
    __slots__ = ("py_value", "info", "c_name", "name", "extras")
    def __init__(self, value: ConstantType, info: str = "",
                 c_name: Optional[str] = None, name: Optional[str] = None,
                 extras: Optional[Dict] = None):
        if not isinstance(value, (int, float, bool, str)):
            raise TypeError(f"Unsupported constant type: {type(value).__name__}")
        self.py_value = value      # keep original type for Python export
        self.info = info
        self.c_name = c_name
        self.name = name
        self.extras = extras or {}
    def __str__(self):
        return f"ConstantSpec(name={self.name!r}, py_value={self.py_value!r}, info={self.info!r}, c_name={self.c_name!r})"
    def __repr__(self):
        return f"<{self} at {hex(id(self))}>"

def constant(value: ConstantType, info: str = "", *,
             c_name: Optional[str] = None, **kwargs) -> ConstantSpec:
    """Declare a numeric constant with optional metadata inside class bodies."""
    return ConstantSpec(value, info, c_name, extras=kwargs)

class GroupSpec:
    __slots__ = ("names", "py_values", "info", "name")
    def __init__(self,
                 names: Tuple[Union[ConstantSpec, GroupSpec, _IntMixin, _FloatMixin, _TupleMixin, str], ...],
                 info: str = "",
                 py_values: Optional[Tuple[ConstantType, ...]] = None,
                 name: Optional[str] = None):
        if not names:
            raise ValueError("group(...) needs at least one constant name")
        self.names = names
        self.py_values = py_values
        self.info = info
        self.name = name
    def __str__(self):
        return f"GroupSpec(name={self.name!r}, names={self.names!r}, py_values={self.py_values!r}, info={self.info!r})"
    def __repr__(self):
        return f"<{self} at {hex(id(self))}>"

def group(*names: Union[ConstantSpec, GroupSpec, Enum, str], info: str = "") -> GroupSpec:
    """Define a named group of constants by their UPPERCASE names (in definition order)."""
    return GroupSpec(tuple(names), info=info)


# ---- Enum Mixin types, to add metadata to constants -------------------------
class _IntMixin(int):
    def __new__(cls, value, info: str = "", c_name: Optional[str] = None, kwargs: Optional[dict]={}):
        return int.__new__(cls, value)
    def __init__(self, value, info: str = "", c_name: Optional[str] = None, kwargs: Optional[dict]={}):
        self.info = info
        self.c_name = c_name
        for kk, vv in kwargs.items():
            setattr(self, kk, vv)
        doc = [self.info] if self.info != '' else []
        c_name = [f"Represented in C as {c_name}."] if c_name is not None else []
        additional = []
        if kwargs:
            extra = [f"{kk}={vv!r}" for kk, vv in kwargs.items()]
            additional.append("Additional info: " + ", ".join(extra) + ".")
        self.__doc__ = ' '.join([*doc, *c_name, *additional])

class _FloatMixin(float):
    def __new__(cls, value, info: str = "", c_name: Optional[str] = None, kwargs: Optional[dict]={}):
        return float.__new__(cls, value)
    def __init__(self, value, info: str = "", c_name: Optional[str] = None, kwargs: Optional[dict]={}):
        self.info = info
        self.c_name = c_name
        for kk, vv in kwargs.items():
            setattr(self, kk, vv)
        doc = [self.info] if self.info != '' else []
        c_name = [f"Represented in C as {c_name}."] if c_name is not None else []
        additional = []
        if kwargs:
            extra = [f"{kk}={vv!r}" for kk, vv in kwargs.items()]
            additional.append("Additional info: " + ", ".join(extra) + ".")
        self.__doc__ = ' '.join([*doc, *c_name, *additional])

class _TupleMixin(tuple):
    def __new__(cls, *args):
        if len(args) < 3:
            raise ValueError("_TupleMixin requires *args, info, names.")
        return tuple.__new__(cls, args[:-2])
    def __init__(self, *args):
        self.info = args[-2]
        self.names = args[-1]
        if len(self.names) != len(self):
            raise ValueError("Tuple names length does not match values length.")
        self.__doc__ = f"{self.info} Represents the following constants: {self.names}."

class _StrMixin(str):
    def __new__(cls, value, info: str = "", c_name: Optional[str] = None, kwargs: Optional[dict] = {}):
        return str.__new__(cls, value)

    def __init__(self, value, info: str = "", c_name: Optional[str] = None, kwargs: Optional[dict] = {}):
        self.info = info
        self.c_name = c_name
        for kk, vv in kwargs.items():
            setattr(self, kk, vv)

        doc = [self.info] if self.info != '' else []
        c_name_line = [f"Represented in C as {c_name}."] if c_name is not None else []
        additional = []
        if kwargs:
            extra = [f"{kk}={vv!r}" for kk, vv in kwargs.items()]
            additional.append("Additional info: " + ", ".join(extra) + ".")
        self.__doc__ = ' '.join([*doc, *c_name_line, *additional])


# ---- enum member builder (returns *members*, not necessarily one enum) ------
def _build_members(module_name: str,
                   items: Mapping[str, ConstantSpec],
                   c_prefix: str,
                   allow_mixed: bool,
                   enum_type_name: str
    ) -> Dict[str, Union[Enum, bool]]:  # Returned Enums are f"{enum_type_name}Int" for ints, f"{enum_type_name}Float" for floats
    """
    Create enum members for ints and floats separately, keep bools as True/False.
    Returns: mapping name -> member (int-enum, float-enum, or True/False).
    """
    # Partition by Python type (bool stays separate to preserve True/False)
    ints: Dict[str, Tuple[int, str, Optional[str]]] = {}
    floats: Dict[str, Tuple[float, str, Optional[str]]] = {}
    bools: Dict[str, Tuple[bool, str, Optional[str]]] = {}
    strings: Dict[str, Tuple[str, str, Optional[str]]] = {}

    for k, spec in items.items():
        if isinstance(spec.py_value, bool):
            bools[k] = (spec.py_value, spec.info, spec.c_name or _cjoin(c_prefix, k), spec.extras)
        elif isinstance(spec.py_value, int):
            ints[k] = (spec.py_value, spec.info, spec.c_name or _cjoin(c_prefix, k), spec.extras)
        elif isinstance(spec.py_value, float):
            floats[k] = (spec.py_value, spec.info, spec.c_name or _cjoin(c_prefix, k), spec.extras)
        elif isinstance(spec.py_value, str):
            strings[k] = (spec.py_value, spec.info, spec.c_name or _cjoin(c_prefix, k), spec.extras)
        else:
            raise TypeError(f"Unsupported type for {k}: {type(spec.py_value).__name__}")

    if not allow_mixed:
        if (ints and floats) or (ints and strings) or (floats and strings):
            raise TypeError("Mixed int/float values are not allowed when _reverse_ == 'unique'.")

    members: Dict[str, object] = {}

    if ints:
        int_enum = Enum(f"{enum_type_name}Int", ints, module=module_name, type=_IntMixin)
        int_enum.__doc__ = ''
        for k in ints:
            members[k] = getattr(int_enum, k)

    if floats:
        float_enum = Enum(f"{enum_type_name}Float", floats, module=module_name, type=_FloatMixin)
        float_enum.__doc__ = ''
        for k in floats:
            members[k] = getattr(float_enum, k)

    if strings:
        str_enum = Enum(f"{enum_type_name}Str", strings, module=module_name, type=_StrMixin)
        str_enum.__doc__ = ''
        for k in strings:
            members[k] = getattr(str_enum, k)

    # Bools exported as plain True/False (metadata lives in *_meta)
    for k, (bv, _, _, _) in bools.items():
        members[k] = bv

    return members


def _build_group_members(module_name: str,
                         items: Mapping[str, GroupSpec],
                         enum_type_name: str
    ) -> Dict[str, Enum]:   # Returned Enums are f"{enum_type_name}Tuple"

    members: Dict[str, object] = {}
    tup_enum = Enum(f"{enum_type_name}Tuple", {k: (*spec.py_values, spec.info, spec.names)
                                            for k, spec in items.items()},
                    module=module_name, type=_TupleMixin)
    tup_enum.__doc__ = ''
    for k in items:
        members[k] = getattr(tup_enum, k)

    return members


# ---- metaclass --------------------------------------------------------------
class _ConstantsMeta(type):
    """
    Single-inheritance constants container.
    Subclass Constants and define UPPERCASE = constant(value, info=..., c_name=...) OR plain int/float/bool.

    Enforces:
      - names unique across the process (all modules),
      - values unique across the process if _reverse_ == "unique".

    Module exports (defining module): ordered, as defined
      - <categories>, <categories>_meta, <categories>_src
      - reverse if requested: unique -> <category>_names, multi -> <category>_by_value

    export_to_module (package root): minimal
      - <categories> (+ <category>_names if unique)
    """
    def __new__(mcls, name, bases, ns, **kw):
        cls = super().__new__(mcls, name, bases, dict(ns))
        if name == "Constants":
            return cls  # Nothing to do for the base class

        if not bases or not bases[0] is Constants:
            raise TypeError(f"{name} must inherit from Constants directly; no further subclassing allowed.")

        # Config from this class only (no inheritance)
        category: str      = (ns.get("_category_",   "constant") or "constant").strip()
        plural: str        = _pluralise(category, ns.get("_plural_", None))
        reverse            = ns.get("_reverse_", None)  # "unique" | "multi" | None
        c_prefix: str      = (ns.get("_c_prefix_", "") or "").strip()
        include_guard: str = ns.get("_include_guard_", f"{_cjoin(c_prefix, category.upper())}_H")
        if reverse not in (None, "unique", "multi"):
            raise ValueError(f"Invalid _reverse_ {reverse!r}; must be None, 'unique', or 'multi'.")
        cls._category_ = category
        cls._plural_ = plural
        cls._reverse_ = reverse
        cls._c_prefix_ = c_prefix
        cls._include_guard_ = include_guard

        # Collect specifications in definition order
        own_specs, own_group_specs = _get_specs(ns, c_prefix)

        # Check consistency across modules
        mod_name = cls.__module__
        _register_type(cls, category, plural, reverse)
        _check_cname_collisions(own_specs, mod_name)
        if reverse == "unique":
            _check_value_uniqueness(own_specs, category, mod_name)

        # Store specs in class
        cls._own_specs_ = own_specs
        for kk, vv in own_specs.items():
            setattr(cls, kk, vv)
        name_to_val = {k: _to_value_key(spec.py_value) for k, spec in own_specs.items()}
        if own_group_specs:
            _resolve_group_specs(own_group_specs, category, mod_name, name_to_val)
            for gname, spec in own_group_specs.items():
                setattr(cls, gname, own_group_specs[gname])
            cls._own_group_specs_ = own_group_specs

        # Class-level views (ordered)
        meta_map = {k: {"value": spec.py_value, "c_name": spec.c_name, "info": spec.info, **spec.extras}
            for k, spec in own_specs.items()}
        if reverse == "unique":
            names_map = {v: k for k, v in name_to_val.items()}
        elif reverse == "multi":
            multi: Dict[Union[int, float], list[str]] = {}
            for k, v in name_to_val.items():
                multi.setdefault(v, []).append(k)
            names_map = {k: tuple(v) for k, v in multi.items()}

        # Store the maps in the class to allow merging when exporting to module
        cls._map  = name_to_val
        cls._meta = meta_map
        if reverse:
            cls._names = names_map
        if own_group_specs:
            cls._groups = {k: spec.py_values for k, spec in own_group_specs.items()}

        # Store C source in class to allow merging when exporting to module
        cls._src = _to_c_header(cls._meta, cls._include_guard_)

        # Export to defining module (ordered, no sorting anywhere) - this is the file where the class is defined
        # Multiple classes of the same category can coexist in the same module; the variables will be joined at the module level
        cls.export_to_module(mod_name, include_groups=True, include_src=True, include_meta=True, import_vars=True)

        return cls

    def export_to_module(cls, module_name: str, include_groups: bool = False,
                         include_src: bool = False, include_meta: bool = False,
                         import_vars: bool = False) -> None:
        """Re-export minimal surface to another module (e.g. package __init__)."""
        if isinstance(module_name, ModuleType):
            mod = module_name
            module_name = mod.__name__
        else:
            mod = sys.modules[module_name]
        category = cls._category_
        plural   = cls._plural_
        reverse  = cls._reverse_
        c_prefix = cls._c_prefix_
        include_guard = cls._include_guard_

        # Build members, preserving Python types
        allow_mixed = (reverse != "unique")
        enum_type_name = ''.join(nn.capitalize() for nn in category.split('_'))
        own_members = _build_members(module_name, cls._own_specs_, c_prefix, allow_mixed, enum_type_name)
        if include_groups and hasattr(cls, "_own_group_specs_"):
            own_group_members = _build_group_members(module_name, cls._own_group_specs_, enum_type_name)
        else:
            own_group_members = {}
        for k, member in {**own_members, **own_group_members}.items():
            _check_uniqueness(k, module_name)
            if import_vars:
                setattr(mod, k, member)

        _merge_map(mod, f"{plural}", cls._map)
        if reverse == "unique":
            _merge_map(mod, f"{category}_names", cls._names)
        elif reverse == "multi":
            _merge_map(mod, f"{category}_names", cls._names, extend_values_when_duplicate=True)
        if include_meta:
            _merge_map(mod, f"{plural}_meta", cls._meta)
        if include_src:
            _merge_c_header(mod, f"{plural}_src", cls._src, include_guard)
        if include_groups and hasattr(cls, "_groups"):
            _merge_map(mod, f"{category}_groups", cls._groups)

        # __all__
        all_list = list(getattr(mod, "__all__", []))
        if import_vars:
            for k in {**own_members, **own_group_members}:
                if k not in all_list:
                    all_list.append(k)
        if f"{plural}" not in all_list:
            all_list.append(f"{plural}")
        if reverse and f"{category}_names" not in all_list:
            all_list.append(f"{category}_names")
        if include_meta and f"{plural}_meta" not in all_list:
            all_list.append(f"{plural}_meta")
        if include_src and f"{plural}_src" not in all_list:
            all_list.append(f"{plural}_src")
        if include_groups and hasattr(cls, "_groups") and f"{category}_groups" not in all_list:
            all_list.append(f"{category}_groups")
        setattr(mod, "__all__", all_list)


class Constants(metaclass=_ConstantsMeta):
    """
    Subclass me and define UPPERCASE = constant(value, info=..., c_name=...) OR plain int/float/bool.

    Configure on the subclass:
        _category_       e.g. "particle_state", "interaction_type", ...
        _plural_         e.g. "particle_states" (optional; auto if omitted)
        _reverse_        "unique" | "multi" | None  (default: None)
        _c_prefix_       e.g. "XC" (no trailing underscore needed)
        _include_guard_  e.g. "XC_PARTICLE_STATES_H" (default derived)
    """
    pass



# copyright ############################### #
# This file is part of the Xcoll Package.   #
# Copyright (c) CERN, 2025.                 #
# ######################################### #

from ..wrapper import BaseWrapper
from .engine import Geant4Engine
from .environment import Geant4Environment
from .reference_masses import geant4_masses_meta


class Geant4Wrapper(BaseWrapper):
    """Wrapper for all Geant4 and BDSIM functions."""

    _engine_cls = Geant4Engine
    _environment_cls = Geant4Environment
    _particle_masses_meta = geant4_masses_meta



# copyright ############################### #
# This file is part of the Xcoll package.   #
# Copyright (c) CERN, 2026.                 #
# ######################################### #

import numpy as np

import xobjects as xo


try:
    import orjson
except (ImportError, ModuleNotFoundError):
    orjson = None
import json as stdjson


def json_load(path):
    try:
        if orjson is None:
            with open(path, "r", encoding="utf-8") as fp:
                return stdjson.load(fp)
        else:
            with open(path, "rb") as fp:
                return orjson.loads(fp.read())
    except FileNotFoundError:
        raise FileNotFoundError(f"File {path} not found.") from None


def json_dump(obj, path, *, indent=None):
    if orjson is None:
        with open(path, "w", encoding="utf-8") as fp:
            stdjson.dump(obj, fp, indent=indent, cls=xo.JEncoder)
    else:
        # orjson ignores `indent`; use OPT_INDENT_2 for pretty
        opt = orjson.OPT_INDENT_2 if indent is not None else 0
        try:
            # Fast path: numeric numpy arrays
            bdata = orjson.dumps(obj, option=opt | orjson.OPT_SERIALIZE_NUMPY)
        except TypeError:
            # Fallback: convert ndarrays (incl. string/object) to lists
            bdata = orjson.dumps(obj, option=opt, default=_orjson_default)
        with open(path, "wb") as fp:
            fp.write(bdata)


def _orjson_default(obj):
    # Mirror JEncoder behaviour
    if isinstance(obj, np.ndarray):
        return obj.tolist()
    if np.issubdtype(type(obj), np.integer):
        return int(obj)
    if np.issubdtype(type(obj), np.floating):
        return float(obj)
    if np.issubdtype(type(obj), np.bool_):
        return bool(obj)
    raise TypeError

# def atomic_write_bytes(path: Path, bdata: bytes):
#     path = Path(path)
#     tmp = path.with_suffix(path.suffix + ".tmp")
#     with open(tmp, "wb") as f:
#         f.write(bdata)
#         f.flush()
#         os.fsync(f.fileno())
#     os.replace(tmp, path)



# copyright ############################### #
# This file is part of the Xcoll package.   #
# Copyright (c) CERN, 2024.                 #
# ######################################### #

from ..xoconstants import Constants, constant, group

class XcollInteractions(Constants):
    _category_ = "interaction" # auto-plural -> "interactions"
    _reverse_  = "unique"      # builds interaction_names
    _c_prefix_ = "XC"

    UNITIALISED                 = constant(0, "NAN  // Do not use")

    ENTER_JAW_L                 = constant(-1, "JI   // point (no children)    Set ds > 0 if entering later")
    ENTER_JAW_R                 = constant(-2, "JI   // point (no children)    Set ds > 0 if entering later")
    EXIT_JAW                    = constant(-3, "JO   // point (no children)")
    ENTER_JAW                   = constant(-4, "JI   // point (no children)    Set ds > 0 if entering later    still here for compatibility")
    ABSORBED                    = constant(1,  "A    // point (no children)    Don't use 0 (is default for unitialised)")
    MULTIPLE_COULOMB_SCATTERING = constant(13, "MCS  // continuous")
    PN_ELASTIC                  = constant(14, "PN   // point (no children)")
    PP_ELASTIC                  = constant(15, "PP   // point (no children)")
    SINGLE_DIFFRACTIVE          = constant(16, "SD   // point (no children)")
    COULOMB                     = constant(17, "C    // point (no children)")
    CHANNELLING                 = constant(100, "CH   // continuous")
    DECHANNELLING               = constant(101, "DCH  // point (no children)")
    VOLUME_REFLECTION_TRANS_CH  = constant(102, "VRCH // point (no children)    Transition region around +-xpcrit")
    VOLUME_REFLECTION           = constant(103, "VR   // point (no children)")
    VOLUME_REFLECTION_TRANS_MCS = constant(104, "VRAM // point (no children)    Transition region around t_P")
    MULTIPLE_COULOMB_TRANS_VR   = constant(105, "AMVR // continuous             Transition region around t_P + 2 xpcrit")
    VOLUME_CAPTURE              = constant(106, "VC   // point (no children)")

    POINT_INTERACTIONS          = group(ENTER_JAW_L, ENTER_JAW_R, EXIT_JAW, ENTER_JAW, ABSORBED, PN_ELASTIC, PP_ELASTIC,
                                        SINGLE_DIFFRACTIVE, COULOMB, DECHANNELLING, VOLUME_REFLECTION_TRANS_CH,
                                        VOLUME_REFLECTION, VOLUME_REFLECTION_TRANS_MCS, VOLUME_CAPTURE,
                                        info="All interactions that happen at a single point.")
    CONTINUOUS_INTERACTIONS     = group(MULTIPLE_COULOMB_SCATTERING, CHANNELLING, MULTIPLE_COULOMB_TRANS_VR,
                                        info="All interactions that happen continuously.")
    CRYSTAL_INTERACTIONS        = group(CHANNELLING, DECHANNELLING, VOLUME_REFLECTION_TRANS_CH, VOLUME_REFLECTION,
                                        VOLUME_REFLECTION_TRANS_MCS, MULTIPLE_COULOMB_TRANS_VR, VOLUME_CAPTURE,
                                        info="All interactions that can happen in a crystal.")

shortcuts = {
    int(ll[0]): ll[1].split('//')[0].strip()
    for ll in [(ii['value'], ii['info']) for ii in interactions_meta.values()]
}



# copyright ############################### #
# This file is part of the Xcoll Package.   #
# Copyright (c) CERN, 2025.                 #
# ######################################### #

from numbers import Number
from contextlib import contextmanager

import xtrack as xt
import xtrack.particles.pdg as pdg


class PhysicsSettingsHelper:
    """Helper class to manage physics settings for scattering routines.
    """

    _return_flags = {'photons': [],
                     'leptons': ['electrons', 'muons', 'tauons', 'neutrinos'],
                     'baryons': ['protons', 'neutrons', 'other_baryons'],
                     'mesons':  ['pions', 'kaons', 'other_mesons'],
                     'ions':    []}
    _global_return_flags = ['all', 'all_charged', 'none']
    _return_modifiers = ['neutral']
    _cut_definitions = ['hadron_lower_momentum', 'photon_lower_momentum', 'electron_lower_momentum']

    def __init__(self, engine):
        with self.__class__._in_constructor(self):
            self._engine = engine
            self._use_cuts = True
            # Set flags to default
            self.return_all = None
            self.hadron_lower_momentum_cut = None
            self.photon_lower_momentum_cut = None
            self.electron_lower_momentum_cut = None
            self.include_showers = None

    @property
    def particle_ref(self):
        if self._engine.particle_ref is None:
            return xt.Particles()
        return self._engine.particle_ref

    @property
    def ref_id(self):
        return self.particle_ref.pdg_id[0]

    @property
    def ref_p0c(self):
        return self.particle_ref.p0c[0]

    @property
    def ref_is_lepton(self):
        return pdg.is_lepton(self.ref_id)

    @property
    def ref_is_proton(self):
        return pdg.is_proton(self.ref_id)

    @property
    def ref_is_ion(self):
        return pdg.is_ion(self.ref_id)

    def update(self):
        # Update settings when they are set to use defaults
        for kk, vv in self.__dict__.items():
            if kk.startswith('_') and not kk.startswith('__') and kk.endswith('_use_default'):
                if vv:
                    prop_name = kk[1:-12]
                    setattr(self, prop_name, None)

    def show(self):
        print("Global return flags:")
        for i, ff in enumerate(self._global_return_flags):
            prefix = "└" if i == len(self._global_return_flags) - 1 else "├"
            val = getattr(self, f'return_{ff}')
            flag = f"{ff.replace('_', ' ')}:"
            print(f"  {prefix} {flag:15} {val}")
        print()
        print("Global return modifiers:")
        for i, ff in enumerate(self._return_modifiers):
            prefix = "└" if i == len(self._return_modifiers) - 1 else "├"
            val = getattr(self, f'return_{ff}')
            flag = f"{ff.replace('_', ' ')}:"
            print(f"  {prefix} {flag:15} {val}")
        print()
        print("Individual return flags:")
        for i, (ff, vv) in enumerate(self._return_flags.items()):
            prefix = "└" if i == len(self._return_flags) - 1 else "├"
            val = getattr(self, f'return_{ff}')
            flag = f"{ff.replace('_', ' ')}:"
            print(f"  {prefix} {flag:15} {val}")
            for j, subff in enumerate(vv):
                subprefix = "└" if j == len(vv) - 1 else "├"
                subval = getattr(self, f'return_{subff}')
                subflag = f"{subff.replace('other_', '').replace('_', ' ')}:"
                preprefix = " " if i == len(self._return_flags) - 1 else "│"
                print(f"  {preprefix}   {subprefix} {subflag:11} {subval}")
        if self._use_cuts:
            print()
            print("Energy cuts [eV]:")
            for i, ff in enumerate(self._cut_definitions):
                prefix = "└" if i == len(self._cut_definitions) - 1 else "├"
                val = getattr(self, f'{ff}_cut')
                flag = f"{ff.replace('_', ' ')}:"
                print(f"  {prefix} {flag:28} {val}")

    @property
    def all_flags(self):
        all_flags  = self._global_return_flags.copy()
        all_flags += self._return_flags.keys()
        all_flags += self._return_modifiers
        all_flags += [fff for ff in self._return_flags.values() for fff in ff]
        return [f'return_{ff}' for ff in all_flags] + [f'{ff}_cut' for ff in self._cut_definitions]


    # =====================
    # === Return groups ===
    # =====================

    @property
    def return_all(self):
        return (self._return_neutral and
                self._return_photons and
                self._return_electrons and
                self._return_muons and
                self._return_tauons and
                self._return_neutrinos and
                self._return_protons and
                self._return_neutrons and
                self._return_other_baryons and
                self._return_pions and
                self._return_kaons and
                self._return_other_mesons and
                self._return_ions)

    @return_all.setter
    def return_all(self, val):
        if val is True:
            self.return_neutral = True
            self.return_photons = True
            self.return_electrons = True
            self.return_muons = True
            self.return_tauons = True
            self.return_neutrinos = True
            self.return_protons = True
            self.return_neutrons = True
            self.return_other_baryons = True
            self.return_pions = True
            self.return_kaons = True
            self.return_other_mesons = True
            self.return_ions = True
        elif val is None or val is False:
            # Default settings
            self.return_neutral = None
            self.return_photons = None
            self.return_electrons = None
            self.return_muons = None
            self.return_tauons = None
            self.return_neutrinos = None
            self.return_protons = None
            self.return_neutrons = None
            self.return_other_baryons = None
            self.return_pions = None
            self.return_kaons = None
            self.return_other_mesons = None
            self.return_ions = None
        else:
            self._engine.stop()
            raise ValueError("`return_all` has to be a boolean!")

    @property
    def return_all_charged(self):
        return (not self._return_neutral and
                self._return_electrons and
                self._return_muons and
                self._return_tauons and
                self._return_protons and
                self._return_other_baryons and
                self._return_pions and
                self._return_kaons and
                self._return_other_mesons and
                self._return_ions)

    @return_all_charged.setter
    def return_all_charged(self, val):
        if val is True:
            self.return_neutral = False
            self.return_photons = False
            self.return_electrons = True
            self.return_muons = True
            self.return_tauons = True
            self.return_neutrinos = False
            self.return_protons = True
            self.return_neutrons = False
            self.return_other_baryons = True
            self.return_pions = True
            self.return_kaons = True
            self.return_other_mesons = True
            self.return_ions = True
        elif val is not None and val is not False:
            self._engine.stop()
            raise ValueError("`return_all_charged` has to be a boolean!")

    @property
    def return_none(self):
        return not (self._return_neutral or
                self._return_photons or
                self._return_electrons or
                self._return_muons or
                self._return_tauons or
                self._return_neutrinos or
                self._return_protons or
                self._return_neutrons or
                self._return_other_baryons or
                self._return_pions or
                self._return_kaons or
                self._return_other_mesons or
                self._return_ions)

    @return_none.setter
    def return_none(self, val):
        if val is True:
            self.return_neutral = False
            self.return_photons = False
            self.return_electrons = False
            self.return_muons = False
            self.return_tauons = False
            self.return_neutrinos = False
            self.return_protons = False
            self.return_neutrons = False
            self.return_other_baryons = False
            self.return_pions = False
            self.return_kaons = False
            self.return_other_mesons = False
            self.return_ions = False
        elif val is None or val is False:
            # Default settings
            self.return_neutral = None
            self.return_photons = None
            self.return_electrons = None
            self.return_muons = None
            self.return_tauons = None
            self.return_neutrinos = None
            self.return_protons = None
            self.return_neutrons = None
            self.return_other_baryons = None
            self.return_pions = None
            self.return_kaons = None
            self.return_other_mesons = None
            self.return_ions = None
        else:
            self._engine.stop()
            raise ValueError("`return_none` has to be a boolean!")

    @property
    def return_neutral(self):
        return self._return_neutral

    @return_neutral.setter
    def return_neutral(self, val):
        if val is None:
            val = False
        if not isinstance(val, bool):
            self._engine.stop()
            raise ValueError("`return_neutral` has to be a boolean!")
        self._return_neutral = val
        # This is a modifier flag; need to update default dependent flags
        self.update()
        if val is False:
            self.return_photons = False
            self.return_neutrinos = False
            self.return_neutrons = False

    @property
    def return_leptons(self):
        ret = [getattr(self, f'return_{attr}') for attr in self._return_flags['leptons']]
        if all(ret):
            return True
        elif not any(ret):
            return False
        else:
            return None

    @return_leptons.setter
    def return_leptons(self, val):
        if not isinstance(val, bool) and val is not None:
            self._engine.stop()
            raise ValueError("`return_leptons` has to be a boolean!")
        for attr in self._return_flags['leptons']:
            setattr(self, f'return_{attr}', val)

    @property
    def return_mesons(self):
        ret = [getattr(self, f'return_{attr}') for attr in self._return_flags['mesons']]
        if all(ret):
            return True
        elif not any(ret):
            return False
        else:
            return None

    @return_mesons.setter
    def return_mesons(self, val):
        if not isinstance(val, bool) and val is not None:
            self._engine.stop()
            raise ValueError("`return_mesons` has to be a boolean!")
        for attr in self._return_flags['mesons']:
            setattr(self, f'return_{attr}', val)

    @property
    def return_baryons(self):
        ret = [getattr(self, f'return_{attr}') for attr in self._return_flags['baryons']]
        if all(ret):
            return True
        elif not any(ret):
            return False
        else:
            return None

    @return_baryons.setter
    def return_baryons(self, val):
        if not isinstance(val, bool) and val is not None:
            self._engine.stop()
            raise ValueError("`return_baryons` has to be a boolean!")
        for attr in self._return_flags['baryons']:
            setattr(self, f'return_{attr}', val)


    # ==========================
    # === Return individuals ===
    # ==========================

    @property
    def return_photons(self):
        return self._return_photons

    @return_photons.setter
    def return_photons(self, val):
        if val is None:
            self._return_photons_use_default = True
            val = self.return_neutral
        else:
            self._return_photons_use_default = False
        if not isinstance(val, bool):
            self._engine.stop()
            raise ValueError("`return_photons` has to be a boolean!")
        self._return_photons = val

    @property
    def return_electrons(self):
        return self._return_electrons

    @return_electrons.setter
    def return_electrons(self, val):
        if val is None:
            self._return_electrons_use_default = True
            val = self.ref_is_lepton
        else:
            self._return_electrons_use_default = False
        if not isinstance(val, bool):
            self._engine.stop()
            raise ValueError("`return_electrons` has to be a boolean!")
        self._return_electrons = val

    @property
    def return_muons(self):
        return self._return_muons

    @return_muons.setter
    def return_muons(self, val):
        if val is None:
            self._return_muons_use_default = True
            val = self.ref_is_lepton
        else:
            self._return_muons_use_default = False
        if not isinstance(val, bool):
            self._engine.stop()
            raise ValueError("`return_muons` has to be a boolean!")
        self._return_muons = val

    @property
    def return_tauons(self):
        return self._return_tauons

    @return_tauons.setter
    def return_tauons(self, val):
        if val is None:
            self._return_tauons_use_default = True
            val = self.ref_is_lepton
        else:
            self._return_tauons_use_default = False
        if not isinstance(val, bool):
            self._engine.stop()
            raise ValueError("`return_tauons` has to be a boolean!")
        self._return_tauons = val

    @property
    def return_neutrinos(self):
        return self._return_neutrinos

    @return_neutrinos.setter
    def return_neutrinos(self, val):
        if val is None:
            self._return_neutrinos_use_default = True
            val = self.ref_is_lepton and self.return_neutral
        else:
            self._return_neutrinos_use_default = False
        if not isinstance(val, bool):
            self._engine.stop()
            raise ValueError("`return_neutrinos` has to be a boolean!")
        self._return_neutrinos = val

    @property
    def return_protons(self):
        return self._return_protons

    @return_protons.setter
    def return_protons(self, val):
        if val is None:
            self._return_protons_use_default = True
            val = self.ref_is_proton or self.ref_is_ion
        else:
            self._return_protons_use_default = False
        if not isinstance(val, bool):
            self._engine.stop()
            raise ValueError("`return_protons` has to be a boolean!")
        self._return_protons = val

    @property
    def return_neutrons(self):
        return self._return_neutrons

    @return_neutrons.setter
    def return_neutrons(self, val):
        if val is None:
            self._return_neutrons_use_default = True
            val = (self.ref_is_proton or self.ref_is_ion) and self.return_neutral
        else:
            self._return_neutrons_use_default = False
        if not isinstance(val, bool):
            self._engine.stop()
            raise ValueError("`return_neutrons` has to be a boolean!")
        self._return_neutrons = val

    @property
    def return_other_baryons(self):
        return self._return_other_baryons

    @return_other_baryons.setter
    def return_other_baryons(self, val):
        if val is None:
            val = False
        if not isinstance(val, bool):
            self._engine.stop()
            raise ValueError("`return_other_baryons` has to be a boolean!")
        self._return_other_baryons = val

    @property
    def return_pions(self):
        return self._return_pions

    @return_pions.setter
    def return_pions(self, val):
        if val is None:
            val = False
        if not isinstance(val, bool):
            self._engine.stop()
            raise ValueError("`return_pions` has to be a boolean!")
        self._return_pions = val

    @property
    def return_kaons(self):
        return self._return_kaons

    @return_kaons.setter
    def return_kaons(self, val):
        if val is None:
            val = False
        if not isinstance(val, bool):
            self._engine.stop()
            raise ValueError("`return_kaons` has to be a boolean!")
        self._return_kaons = val

    @property
    def return_other_mesons(self):
        return self._return_other_mesons

    @return_other_mesons.setter
    def return_other_mesons(self, val):
        if val is None:
            val = False
        if not isinstance(val, bool):
            self._engine.stop()
            raise ValueError("`return_other_mesons` has to be a boolean!")
        self._return_other_mesons = val

    @property
    def return_ions(self):
        return self._return_ions

    @return_ions.setter
    def return_ions(self, val):
        if val is None:
            self._return_ions_use_default = True
            val = self.ref_is_ion
        else:
            self._return_ions_use_default = False
        if not isinstance(val, bool):
            self._engine.stop()
            raise ValueError("`return_ions` has to be a boolean!")
        self._return_ions = val


    # =====================
    # === Momentum cuts ===
    # =====================

    @property
    def hadron_lower_momentum_cut(self):
        if not self._use_cuts:
            return None
        return self._hadron_lower_momentum_cut

    @hadron_lower_momentum_cut.setter
    def hadron_lower_momentum_cut(self, val):
        if not self._use_cuts:
            self._hadron_lower_momentum_cut = None
            return
        self._hadron_lower_momentum_cut_use_default = False
        if val is None:
            self._hadron_lower_momentum_cut_use_default = True
            val = self.ref_p0c / 10
            if self.ref_is_ion:
                _, A, _, _ = pdg.get_properties_from_pdg_id(self.ref_id)
                val /= A
        elif not isinstance(val, Number) or val < 0:
            self._engine.stop()
            raise ValueError("`hadron_lower_momentum_cut` has to be a non-negative number!")
        elif val < 1.e9:
            self._print(f"Warning: Hadron lower momentum cut of {val/1.e9}GeV "
                       + "is very low and will result in very long computation times.")
        self._hadron_lower_momentum_cut = val

    @property
    def photon_lower_momentum_cut(self):
        if not self._use_cuts:
            return None
        return self._photon_lower_momentum_cut

    @photon_lower_momentum_cut.setter
    def photon_lower_momentum_cut(self, val):
        if not self._use_cuts:
            self._photon_lower_momentum_cut = None
            return
        self._photon_lower_momentum_cut_use_default = False
        if val is None:
            self._photon_lower_momentum_cut_use_default = True
            val = self.ref_p0c * 1e-3
        elif not isinstance(val, Number) or val < 0:
            self._engine.stop()
            raise ValueError("`photon_lower_momentum_cut` has to be a non-negative number!")
        elif val < 1.e3:
            self._print(f"Warning: Photon lower momentum cut of {val/1.e3}keV "
                       + "is very low and will result in very long computation times.")
        self._photon_lower_momentum_cut = val

    @property
    def electron_lower_momentum_cut(self):
        if not self._use_cuts:
            return None
        return self._electron_lower_momentum_cut

    @electron_lower_momentum_cut.setter
    def electron_lower_momentum_cut(self, val):
        if not self._use_cuts:
            self._electron_lower_momentum_cut = None
            return
        self._electron_lower_momentum_cut_use_default = False
        if val is None:
            self._electron_lower_momentum_cut_use_default = True
            if self.ref_is_lepton:
                val = self.ref_p0c / 10
            else:
                val = self.ref_p0c * 1e-3
        elif not isinstance(val, Number) or val < 0:
            self._engine.stop()
            raise ValueError("`electron_lower_momentum_cut` has to be a non-negative number!")
        elif val < 1.e6:
            self._print(f"Warning: Electron lower momentum cut of {val/1.e6}MeV "
                       + "is very low and will result in very long computation times.")
        self._electron_lower_momentum_cut = val

    @property
    def include_showers(self):
        if not self._use_cuts:
            return None
        return self._include_showers

    @include_showers.setter
    def include_showers(self, val):
        if not self._use_cuts:
            self._include_showers = None
            return
        if val is None:
            self._include_showers_use_default = True
            val = True if self.ref_is_lepton else False
            val = True if self.return_all else val
        else:
            self._include_showers_use_default = False
        if not isinstance(val, bool):
            self._engine.stop()
            raise ValueError("`include_showers` has to be a boolean!")
        self._include_showers = val


    def __getattribute__(self, item):
        # always use base lookup inside this method
        obj_get = object.__getattribute__

        if item.startswith("return_"):
            # compute flags WITHOUT going through self.<property>
            global_flags     = obj_get(self, "_global_return_flags")
            return_flags     = obj_get(self, "_return_flags")
            return_modifiers = obj_get(self, "_return_modifiers")

            all_flags = list(global_flags)
            all_flags += list(return_flags.keys())
            all_flags += list(return_modifiers)
            all_flags += [fff for ff in return_flags.values() for fff in ff]

            if item not in {f"return_{ff}" for ff in all_flags}:
                engine = obj_get(self, "_engine") # also fetch engine via base lookup to avoid re-entry surprises
                engine.stop()
                raise AttributeError(f"Return flag '{item}' does not exist!")

        elif not item.startswith('_') and item.endswith("_cut"):
            cut_defs = obj_get(self, "_cut_definitions")
            if item not in {f"{ff}_cut" for ff in cut_defs}:
                engine = obj_get(self, "_engine")
                engine.stop()
                raise AttributeError(f"Cut definition '{item}' does not exist!")

        return obj_get(self, item)

    def __setattr__(self, name, value):
        if not hasattr(self, name):
            raise AttributeError(f"'{self.__class__.__name__}' object has no attribute '{name}'")
        super().__setattr__(name, value)

    @classmethod
    @contextmanager
    def _in_constructor(cls, self=None):
        original_setattr = cls.__setattr__
        def new_setattr(self, *args, **kwargs):
            return super().__setattr__( *args, **kwargs)
        cls.__setattr__ = new_setattr
        try:
            yield
        finally:
            cls.__setattr__ = original_setattr